from unittest.mock import patch

from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid
import grid.commands.credentials as credentials

RUNNER = CliRunner()


class TestInteractive:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        credentials.gql = lambda x: x
        grid.gql = lambda x: x

        cls.mp_status = lambda *args, **kwargs: True
        cls.CREDENTIAL_ID = 'test-credential'

    def test_interactive_succeeds_with_no_args(self, monkeypatch):
        """
        grid interactive succeeds with no args
        """
        monkeypatch.setattr(grid.Grid, 'status', self.mp_status)
        result = RUNNER.invoke(cli.interactive, [])
        assert result.exit_code == 0
        assert not result.exception

    def test_interactive_create_succeeds(self, monkeypatch):
        """
        grid interactive create succeeds
        """
        def mp_create_interactive_node(*args, **kwargs):
            return {'success': True, 'message': ''}

        monkeypatch.setattr(grid.Grid, 'status', self.mp_status)
        monkeypatch.setattr(resolvers, 'create_interactive_node', mp_create_interactive_node)

        result = RUNNER.invoke(cli.interactive, ["create"])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Interactive node created!' in result.output

    def test_interactive_create_fails_multiple_creds(self, monkeypatch):
        """
        grid interactive create fails with multiple non-default creds
        """
        creds = self.CREDENTIAL_ID

        def mp_get_credentials_multiple_no_default(self):
            return {
                'getUserCredentials': [{
                    'credentialId': creds,
                    'provider': 'aws',
                    'defaultCredential': False,
                    'createdAt': '2020-11-11',
                    'lastUsedAt': '2020-11-11',
                    'alias': None
                }, {
                    'credentialId': "Another",
                    'provider': 'aws',
                    'defaultCredential': False,
                    'createdAt': '2020-11-11',
                    'lastUsedAt': '2020-11-11',
                    'alias': None
                }]
            }

        monkeypatch.setattr(grid.Grid, 'get_credentials', mp_get_credentials_multiple_no_default)
        result = RUNNER.invoke(cli.interactive, ["create"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Detected multiple credentials. Which would you like to use?' in result.output

    def test_interactive_create_fails_with_wrong_creds(self, monkeypatch):
        """
        grid interactive create fails with wrong credentials
        """
        def mp_no_default_creds(self):
            return {
                'getUserCredentials': [{
                    'credentialId': 'wrong-creds',
                    'provider': 'aws',
                    'defaultCredential': True
                }]
            }

        monkeypatch.setattr(grid.Grid, 'get_credentials', mp_no_default_creds)
        result = RUNNER.invoke(cli.interactive, ["create", '--grid_credential', self.CREDENTIAL_ID])
        assert result.exit_code == 1
        assert result.exception
        assert f'Credential ID {self.CREDENTIAL_ID} does not exist.' in result.output

    @staticmethod
    def test_interactive_create_fails_with_no_creds(monkeypatch):
        """
        grid interactive create fails with no credentials
        """
        def mp_no_creds(self):
            return {'getUserCredentials': []}

        monkeypatch.setattr(grid.Grid, 'get_credentials', mp_no_creds)
        result = RUNNER.invoke(cli.interactive, ["create"])
        assert result.exit_code == 1
        assert result.exception
        assert ' No cloud credentials available' in result.output

    def test_interactive_create_succeeds_with_default_creds(self, monkeypatch):
        """
        grid interactive create succeeds with default creds
        """
        creds = self.CREDENTIAL_ID

        def mp_default_creds(self):
            return {'getUserCredentials': [{'credentialId': creds, 'provider': 'aws', 'defaultCredential': True}]}

        monkeypatch.setattr(grid.Grid, 'get_credentials', mp_default_creds)
        result = RUNNER.invoke(cli.interactive, ["create"])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Interactive node created!' in result.output

    @staticmethod
    def test_interactive_create_fails_if_config_not_found():
        """grid interactive create fails if config not found"""

        result = RUNNER.invoke(cli.interactive, ["create", "--grid_config", "no-file"])

        assert result.exit_code == 2
        assert result.exception
        assert 'Could not open file' in result.output

    def test_interactive_create_succeeds_with_config(self, tmp_path, monkeypatch):
        """grid interactive create succeeds with config"""
        content = """
        compute:
            provider:
                credentials: cc-kmz2f
                region: us-east-1
                vendor: aws
            train:
                disk_size: 200
                gpus: 0
                instance: g4dn.xlarge
                max_nodes: 10
                nodes: 0
                scale_down_seconds: 30
                memory: 100Mi
                cpus: 2
        """

        config = tmp_path / "config.yaml"
        config.write_text(content)

        monkeypatch.setattr(grid.Grid, 'status', self.mp_status)
        result = RUNNER.invoke(cli.interactive, ["create", "--grid_config", config.as_posix()])

        assert result.exit_code == 0
        assert not result.exception
        assert 'Interactive node created' in result.output

    @staticmethod
    def test_interactive_delete_succeeds(monkeypatch):
        """grid interactive delete succeeds"""
        def mp_delete_interactive_node(*args, **kwargs):
            return {'success': True, 'message': ''}

        monkeypatch.setattr(resolvers, 'delete_interactive_node', mp_delete_interactive_node)
        result = RUNNER.invoke(cli.interactive, ["delete", "node-to-delete"])

        assert result.exit_code == 0
        assert not result.exception
        assert 'has been deleted successfully' in result.output

    @staticmethod
    def test_interactive_delete_fails(monkeypatch):
        """grid interactive delete fails"""
        def mp_delete_interactive_node(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'delete_interactive_node', mp_delete_interactive_node)
        result = RUNNER.invoke(cli.interactive, ["delete", "node-to-delete"])

        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to delete interactive node' in result.output

    def test_interactive_create_description(self, monkeypatch):
        """
        grid interactive create with description
        """
        def mp_create_interactive_node(*args, **kwargs):
            return {'success': True, 'message': ''}

        monkeypatch.setattr(grid.Grid, 'status', self.mp_status)
        monkeypatch.setattr(resolvers, 'create_interactive_node', mp_create_interactive_node)

        result = RUNNER.invoke(cli.interactive, ["create", "--grid_description", "test-description"])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Interactive node created!' in result.output

    @patch("grid.client.Grid.sync_ssh_config",
           autospec=True,
           return_value=[
               {
                   "name": "test-pending",
                   "ssh_url": "ssh://jovyan@ix2.grid.ai:433",
                   "status": "pending",
               },
           ])
    def test_update_ssh_config_and_check_ixsession_status_callback(self, ixsessions):
        result = RUNNER.invoke(cli.interactive, ["ssh", "not-exist-test"])
        assert result.exit_code == 2
        assert result.output
        assert "does not exist" in result.output

        result = RUNNER.invoke(cli.interactive, ["ssh", "test-pending"])
        assert result.exit_code == 2
        assert result.output
        assert "not ready" in result.output

    def test_session_with_nickname(self, monkeypatch):
        """
        grid interactive create succeeds with nickname
        """
        def mp_create_interactive_node(*args, **kwargs):
            return {'success': True, 'message': ''}

        monkeypatch.setattr(grid.Grid, 'status', self.mp_status)
        monkeypatch.setattr(resolvers, 'create_interactive_node', mp_create_interactive_node)

        result = RUNNER.invoke(cli.interactive, ["create", '--grid_instance_type', '2_cpu_4gb'])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Interactive node created!' in result.output



import csv
import glob
import json
import os
from pathlib import Path
from unittest.mock import MagicMock
from urllib.parse import urljoin

import click
from click.exceptions import ClickException
from gql import Client
from gql.transport.exceptions import TransportQueryError
import httpretty
import pytest
import requests
from tests.mock_backend import GridAIBackenedTestServer
from tests.mock_backend import resolvers

import grid.client as grid
from grid.datastore import DatastoreUploadSession
from grid.exceptions import AuthenticationError
from grid.exceptions import TrainError
import grid.globals as env
from grid.metadata import __version__
from grid.types import ObservableType
from grid.types import WorkflowType

#  Prevents Click from opening the browser.
click.launch = lambda x: True


def monkey_patch_observables():
    """Monkey patches the observables factory."""
    class MonkeyPatchedObservable:
        key = 'getRuns'

        def __init__(self, *args, **kwargs):  # skipcq: PTC-W0049
            pass

        def get(self, *args, **kwarrgs):
            return {self.key: [{'columnn': 'value'}, {'column_fail_a': []}, {'column_fail_b': {}}]}

    return {
        ObservableType.EXPERIMENT: MonkeyPatchedObservable,
        ObservableType.RUN: MonkeyPatchedObservable,
        ObservableType.CLUSTER: MonkeyPatchedObservable,
    }


#  Test authentication errors
def monkey_patch_execute(*args, **kwargs):
    raise requests.exceptions.HTTPError('error')


class MonkeyPatchClient:
    def execute(self, *args, **kwargs):
        raise Exception("{'message': 'test exception'}")


class TestGridClient:
    """Unit tests for the Grid class."""
    @classmethod
    def setup_class(cls):
        #  Monkey patches the GraphQL client to read from a local schema.
        def monkey_patch_client(self):
            self.client = GridAIBackenedTestServer()

        #  Monkey-patches the gql method so that it passes
        #  forward a GraphQL query string directly.
        grid.gql = lambda x: x

        #  skipcq: PYL-W0212
        grid.Grid._init_client = monkey_patch_client

        cls.creds_path = 'tests/data/credentials.json'
        cls.creds_path_incorrect = 'tests/data/test-credentials.json'
        cls.grid_header_keys = ['X-Grid-User', 'X-Grid-Key']

        cls.train_kwargs = {
            'config':
            'test-config',
            'kind':
            WorkflowType.SCRIPT,
            'run_name':
            'test-run',
            'run_description':
            'test description',
            'entrypoint':
            'test_file.py',
            'script_args': ['--learning_rate', '0.001'],
            'invocation_command':
            'grid train --grid_name test-run --grid_description "test description" test_file.py --learning_rate 0.001'
        }

    def remove_env(self):
        #  Makes sure that the Grid env vars are not set
        envs_to_remove = ['GRID_CREDENTIAL_PATH', 'GRID_USER_ID', 'GRID_API_KEY']
        for env in envs_to_remove:
            if os.getenv(env):
                del os.environ[env]

    def remove_status_files(self):
        # path = 'test/data/'
        for e in ['csv', 'json']:
            for f in glob.glob(f'*.{e}'):
                os.remove(f)

    def remove_credentials_file(self):
        p = Path(self.creds_path_incorrect)
        if p.exists():
            p.unlink(missing_ok=True)

    def setUp(self):
        self.remove_env()
        self.remove_status_files()
        self.remove_credentials_file()

    def teardown(self):
        self.remove_env()

        #  Removes test credentials added to home path.
        P = Path.home().joinpath(self.creds_path)
        if P.exists():
            P.unlink(missing_ok=True)

        self.remove_status_files()
        self.remove_credentials_file()

    def test_client_local_path(self):
        """Client with local credentials path initializes correctly"""
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        for key in self.grid_header_keys:
            assert key in G.headers.keys()

    def test_client_loads_credentials_from_env_var(self):
        """Client loads credentials path from env var"""
        os.environ['GRID_CREDENTIAL_PATH'] = self.creds_path
        G = grid.Grid()
        for key in self.grid_header_keys:
            assert key in G.headers.keys()

        os.environ['GRID_CREDENTIAL_PATH'] = 'fake-path'
        with pytest.raises(click.ClickException):
            grid.Grid()

        self.remove_env()

    def test_client_raises_exception_if_creds_path_not_found(self):
        """Client raises exception if credentials path not found"""
        credentials_path = 'tests/data/foo.json'
        with pytest.raises(click.ClickException):
            grid.Grid(credential_path=credentials_path)

    def test_nested_path_is_parsed_correctly(self):
        """Tests that we can add the Git root path to a script"""
        result = grid.Grid._add_git_root_path(entrypoint='foo.py')
        path_elems = result.split(os.path.sep)

        actual = path_elems[-2:]
        expected = ['grid-cli', 'foo.py']
        assert len(actual) == len(expected)
        assert all(a == b for a, b in zip(actual, expected))

    def test_client_loads_credentials_from_default_path(self):
        """Client loads credentials from default path"""
        test_path = 'tests/data/credentials.json'
        with open(test_path) as f:
            credentials = json.load(f)

        #  Let's create a credentials file in the home
        #  directory.
        creds_name = 'test_credentials.json'
        P = Path.home().joinpath(creds_name)
        with P.open('w') as f:
            json.dump(credentials, f)

        grid.Grid.grid_credentials_path = creds_name
        G = grid.Grid()

        assert G.credentials.get('UserID') == credentials['UserID']

    def test_client_raises_error_if_no_creds_available(self):
        """Client loads credentials from default path"""
        with pytest.raises(click.ClickException):
            grid.Grid(credential_path=self.creds_path_incorrect)

        # TODO: test not working
        # with pytest.raises(click.ClickException):
        #     # monkeypatch.setattr(grid.Grid, 'grid_credentials_path', self.creds_path_incorrect):
        #     grid.Grid.grid_credentials_path = self.creds_path_incorrect
        #     G = grid.Grid(load_local_credentials=False)
        #     G._set_local_credentials()

    #  NOTE: there's a race condition here with the env
    #  var. Let's leave this named this way.
    def test_a_client_local_init(self):
        """Client init without local credentials leaves headers unchanged"""
        assert not os.getenv('GRID_CREDENTIAL_PATH')

        G = grid.Grid(load_local_credentials=False)
        for key in self.grid_header_keys:
            assert key not in G.headers.keys()

    def test_train(self, monkeypatch):
        """grid.Grid().train() executes a training operation correctly."""
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        env.IGNORE_WARNINGS = True
        G.train(**self.train_kwargs)

        # We want to ignore global settings to test debugging.
        monkeypatch.setattr(grid.Grid, '_load_global_settings', lambda: True)
        env.DEBUG = True
        G.train(**self.train_kwargs)

    def test_train_raises_exception_github_not_accessible(self, monkeypatch):
        """grid.Grid().train() raises exception if the user did not authorize
        access to a private repo.
        """
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        env.IGNORE_WARNINGS = True

        def monkey_patch_is_github_repository_accessible(*args, **kwargs):
            return {
                "isAccessible": False,
            }

        with pytest.raises(click.exceptions.ClickException) as ex_info:
            monkeypatch.setattr(resolvers, 'is_github_repository_accessible',
                                monkey_patch_is_github_repository_accessible)
            G.train(**self.train_kwargs)

        assert "Please grant access on the settings page" in ex_info.value.args[0]

    def test_train_raises_exception_blueprint(self):
        """
        grid.Grid().train() raises exception when attempting to
        train a blueprint.
        """
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        env.IGNORE_WARNINGS = True
        with pytest.raises(TrainError):
            G.train(**{**self.train_kwargs, 'kind': WorkflowType.BLUEPRINT})

    def test_train_raises_exception_if_query_fails(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True
        G._check_github_repo_accessible = lambda _: True
        env.IGNORE_WARNINGS = True

        with pytest.raises(click.ClickException):
            G.train(**self.train_kwargs)

    def test_status_returns_results(self):
        """grid.Grid().status() returns a dict results."""
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.available_observables = monkey_patch_observables()
        results = G.status()
        assert isinstance(results, dict)

    def test_status_generates_output_files(self):
        """grid.Grid().status() generates output files."""
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.available_observables = monkey_patch_observables()

        #  Tests exporting.
        extensions = ['csv', 'json']
        for e in extensions:
            G.status(export=e)
            files = [*glob.glob(f'*.{e}')]
            assert len(files) == 1

            #  Test that lists or dict columns are not exported to CSV.
            if e == 'csv':
                with open(files[0], 'r') as f:
                    data = [*csv.DictReader(f)]
                    assert 'column_fail_a' not in data[0].keys()
                    assert 'column_fail_b' not in data[0].keys()

    def test_user_id(self):
        """Returns same User ID as specified in credentials file."""
        test_path = 'tests/data/credentials.json'
        with open(test_path) as f:
            credentials = json.load(f)

        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        assert G.credentials.get('UserID') == credentials['UserID']
        assert G.user_id == credentials['UserID']

    def test_setting_local_credentials_using_env_vars(self):
        """grid.Grid() sets credentials using environment variables."""
        test_user_id = 'test-user-id'
        test_api_key = 'test-api_key'

        os.environ['GRID_USER_ID'] = test_user_id
        os.environ['GRID_API_KEY'] = test_api_key
        G = grid.Grid(load_local_credentials=True)

        assert G.headers['X-Grid-User'] == test_user_id
        assert G.headers['X-Grid-Key'] == test_api_key

        self.remove_env()

    @staticmethod
    def test_check_user_github_token():
        """
        grid.Grid()._check_user_github_token() checks if user's GH
        token is valid.
        """
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G._check_user_github_token()
        assert result is True

    @staticmethod
    def test_check_user_github_token_raises_exception(monkeypatch):
        """
        grid.Grid()._check_user_github_token() raises exception if no
        Github token is available.
        """
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        #  Test when users don't have a valid token
        def monkey_patch_token(*args, **kwargs):
            return {'hasValidToken': False}

        with pytest.raises(click.exceptions.ClickException):
            monkeypatch.setattr(resolvers, 'check_user_github_token', monkey_patch_token)
            G._check_user_github_token()

        with pytest.raises(AuthenticationError):
            monkeypatch.setattr(GridAIBackenedTestServer, 'execute', monkey_patch_execute)
            # Prevents browser from opening
            monkeypatch.setattr(click, 'launch', lambda x: True)
            G._check_user_github_token()

    @staticmethod
    def test_add_git_root_path():
        """Grid._add_git_root_path() adds the git root path to script."""
        G = grid.Grid(load_local_credentials=False)
        result = G._add_git_root_path(entrypoint='test.py')

        assert result == '/grid-cli/test.py'

    @staticmethod
    def test_login():
        """Grid.login() correctly sends login query."""
        G = grid.Grid(load_local_credentials=False)

        G.login(username='test-user', key='test-key')

    @staticmethod
    def test_login_handles_error(monkeypatch):
        """Grid.login() handles HTTPError exception."""
        G = grid.Grid(load_local_credentials=False)

        with pytest.raises(ClickException):
            monkeypatch.setattr(GridAIBackenedTestServer, 'execute', monkey_patch_execute)
            G.login(username='test-user', key='test-key')

    @staticmethod
    def test_validate_datastore_version_exception(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        env.DEBUG = True

        def monkey_path_get_datastores_exc(*args, **kwargs):
            raise TransportQueryError(msg="")

        config = {"compute": {"train": {"datastore_name": "mnist", "datastore_version": None}}}
        with pytest.raises(click.ClickException) as ex_info:
            monkeypatch.setattr(GridAIBackenedTestServer, 'execute', monkey_path_get_datastores_exc)
            G.validate_datastore_version(grid_config=config)
        assert "Unable to list datastores" in ex_info.value.args[0]

    @staticmethod
    def test_validate_datastore_version(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        env.DEBUG = True

        def monkey_path_get_datastores(*args, **kwargs):
            return [
                {
                    'name': 'mnist',
                    'version': '1',
                    'snapshotStatus': 'succeeded'
                },
                {
                    'name': 'mnist',
                    'version': '2',
                    'snapshotStatus': 'succeeded'
                },
                {
                    'name': 'mnist',
                    'version': '3',
                    'snapshotStatus': 'unknown'
                },
            ]

        monkeypatch.setattr(resolvers, 'get_datastores', monkey_path_get_datastores)
        config = {}
        G.validate_datastore_version(grid_config=config)
        assert config == {}

        config = {
            "compute": {
                "train": {
                    "datastore_name": "mnist",
                }
            }
        }
        G.validate_datastore_version(grid_config=config)
        assert config["compute"]["train"]["datastore_name"] == "mnist"
        assert "datastore_version" not in config["compute"]["train"]

        config = {"compute": {"train": {"datastore_name": "mnist", "datastore_version": None}}}
        G.validate_datastore_version(grid_config=config)
        assert config["compute"]["train"]["datastore_name"] == "mnist"
        assert config["compute"]["train"]["datastore_version"] == "2"

        config = {"compute": {"train": {"datastore_name": "mnist", "datastore_version": ""}}}
        G.validate_datastore_version(grid_config=config)
        assert config["compute"]["train"]["datastore_name"] == "mnist"
        assert config["compute"]["train"]["datastore_version"] == "2"

        config = {"compute": {"train": {"datastore_name": "mnist", "datastore_version": "1"}}}
        G.validate_datastore_version(grid_config=config)
        assert config["compute"]["train"]["datastore_name"] == "mnist"
        assert config["compute"]["train"]["datastore_version"] == "1"

        config = {"compute": {"train": {"datastore_name": "wrong-name", "datastore_version": None}}}
        with pytest.raises(click.ClickException) as ex_info:
            G.validate_datastore_version(grid_config=config)
        assert "unable to find a ready-to-use version" in ex_info.value.args[0]

        config = {"compute": {"train": {"datastore_name": "wrong-name", "datastore_version": "1"}}}
        G.validate_datastore_version(grid_config=config)
        assert config["compute"]["train"]["datastore_name"] == "wrong-name"
        assert config["compute"]["train"]["datastore_version"] == "1"

    @staticmethod
    def test_upload_datstore_handles_error():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()

        with pytest.raises(click.ClickException):
            G.upload_datastore(source='./tests/data/datastore_test/',
                               name='datastore-test',
                               credential_id='test-credential',
                               compression=True)

    @staticmethod
    def test_resuming_upload_datastore():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()

        session = DatastoreUploadSession(name="test_session", version=1, source_dir="data", credential_id="cc-abcdef")
        upload = MagicMock()
        session.upload = upload

        DatastoreUploadSession.recover_sessions = MagicMock(return_value=[session])
        G.resume_datastore_session("test_session-v1")
        upload.assert_called_once()

    @staticmethod
    def test_experiment_details():
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.experiment_details(experiment_id='test-experiment-id')
        assert result['getExperimentDetails']['status'] == 'succeeded'

    @staticmethod
    def test_create_interactive_node(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.create_interactive_node(name='test-interactive-id', config='', description="test-description")
        assert result

        def monkey_path_create(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'create_interactive_node', monkey_path_create)
        env.DEBUG = True
        with pytest.raises(click.ClickException):
            G.create_interactive_node(name='test-interactive-id', config='', description='test-description')

    @staticmethod
    def test_create_interactive_node_handles_error():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True

        with pytest.raises(click.ClickException):
            G.create_interactive_node(name='test-interactive-id', config='', description="test-desc")

    @staticmethod
    def test_pause_interactive_node(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.pause_interactive_node(interactive_node_name='test-interactive-id')
        assert result

        def monkey_path_pause(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'pause_interactive_node', monkey_path_pause)
        env.DEBUG = True
        with pytest.raises(click.ClickException):
            G.pause_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_pause_interactive_node_handles_error():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True

        with pytest.raises(click.ClickException):
            G.pause_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_resume_interactive_node(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.resume_interactive_node(interactive_node_name='test-interactive-id')
        assert result

        def monkey_path_pause(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'resume_interactive_node', monkey_path_pause)
        env.DEBUG = True
        with pytest.raises(click.ClickException):
            G.resume_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_resume_interactive_node_handles_error():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True

        with pytest.raises(click.ClickException):
            G.resume_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_delete_interactive_node(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.delete_interactive_node(interactive_node_name='test-interactive-id')
        assert result

        def monkey_path_delete(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'delete_interactive_node', monkey_path_delete)
        env.DEBUG = True
        with pytest.raises(click.ClickException):
            G.delete_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_delete_interactive_node_handles_error():
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True

        with pytest.raises(click.ClickException):
            G.delete_interactive_node(interactive_node_name='test-interactive-id')

    @staticmethod
    def test_get_experiment_id(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.get_experiment_id(experiment_name='some-name')
        assert result == 'some-id'

        def mp_get_experiment_id(*args, **kwargs):
            return {'success': False, 'message': ''}

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        with pytest.raises(click.ClickException) as ex_info:
            G.get_experiment_id(experiment_name='some-name')
        assert "Could not find experiment" in ex_info.value.args[0]

        def mp_get_experiment_id(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        with pytest.raises(click.ClickException) as ex_info:
            G.get_experiment_id(experiment_name='some-name')
        assert "Could not find experiment" in ex_info.value.args[0]

    @staticmethod
    def test_delete():
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.delete(experiment_name='test-experiment-name')
        assert result

        result = G.delete(run_name='test-run-name')
        assert result

    @staticmethod
    def test_delete_handles_errors(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G._init_client()
        G._check_user_github_token = lambda: True

        def mp_delete_object(**kwargs):
            return {'success': False, 'message': None}

        monkeypatch.setattr(resolvers, 'delete_experiment', mp_delete_object)
        with pytest.raises(click.ClickException):
            G.delete(experiment_name='test-experiment-name')

        G.client = MonkeyPatchClient()
        with pytest.raises(click.ClickException):
            G.delete(experiment_name='test-experiment-name')

        monkeypatch.setattr(resolvers, 'delete_run', mp_delete_object)
        with pytest.raises(click.ClickException):
            G.delete(experiment_name='test-run-name')

        G.client = MonkeyPatchClient()
        with pytest.raises(click.ClickException):
            G.delete(run_name='test-run-name')

    @staticmethod
    def test_cancel():
        G = grid.Grid(load_local_credentials=False)
        G._init_client()

        result = G.cancel(run_name='test-run-name')
        assert result is True

        result = G.cancel(experiment_name='test-experiment-name')
        assert result is True

    @staticmethod
    def test_cancel_handles_errors(monkeypatch):
        G = grid.Grid(load_local_credentials=False)
        G.client = MonkeyPatchClient()
        G._check_user_github_token = lambda: True

        def monkey_patch_experiment_details(**kwargs):
            return 'failed'

        monkeypatch.setattr(G, 'experiment_details', monkey_patch_experiment_details)
        with pytest.raises(click.ClickException):
            G.cancel(run_name='test-run-id')

    def test_history(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        gql = MagicMock(Client, autospec=True)
        G.client = gql

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.EXPERIMENT] = observable_fact
        observable.get_history = MagicMock(return_value={"a": "b"})
        assert G.history(
            kind=ObservableType.EXPERIMENT,
            identifiers=["foo"],
        ) == {
            "a": "b"
        }
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client, identifier="foo")
        observable.get_history.assert_called_once()

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.RUN] = observable_fact
        observable.get_history = MagicMock(return_value={"a": "b"})
        assert G.history(kind=ObservableType.RUN, ) == {"a": "b"}
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client)
        observable.get_history.assert_called_once()

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.RUN] = observable_fact
        observable.get_history = MagicMock(return_value={"a": "b"})
        assert G.history() == {"a": "b"}
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client)
        observable.get_history.assert_called_once()

        with pytest.raises(click.BadArgumentUsage):
            G.history(kind=ObservableType.INTERACTIVE)

        with pytest.raises(click.BadArgumentUsage):
            G.history(kind=ObservableType.CLUSTER)

    def test_status(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)
        gql = MagicMock(Client, autospec=True)
        G.client = gql

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.EXPERIMENT] = observable_fact
        observable.get = MagicMock(return_value={"a": "b"})
        assert G.status(
            kind=ObservableType.EXPERIMENT,
            identifiers=["foo"],
        ) == {
            "a": "b"
        }
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client, identifier="foo")
        observable.get.assert_called_once()

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.RUN] = observable_fact
        observable.get = MagicMock(return_value={"a": "b"})
        assert G.status(kind=ObservableType.RUN, ) == {"a": "b"}
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client)
        observable.get.assert_called_once()

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.RUN] = observable_fact
        observable.get = MagicMock(return_value={"a": "b"})
        assert G.status() == {"a": "b"}
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client)
        observable.get.assert_called_once()

        observable = MagicMock()
        observable_fact = MagicMock(return_value=observable)
        G.available_observables[ObservableType.INTERACTIVE] = observable_fact
        observable.get = MagicMock(return_value={"a": "b"})
        assert G.status(kind=ObservableType.INTERACTIVE) == {"a": "b"}
        observable_fact.assert_called_once()
        observable_fact.assert_called_once_with(client=G.client)
        observable.get.assert_called_once()

        with pytest.raises(click.BadArgumentUsage):
            G.status(kind=ObservableType.CLUSTER)

    def test_experiment_logs_queued(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock()
        G.get_experiment_id = MagicMock()
        G.experiment_details = MagicMock(
            return_value={'getExperimentDetails': {
                'status': 'queued',
                'experimentId': 'test'
            }})
        G.experiment_logs('test')
        G.experiment_details.assert_called_once()
        G.get_experiment_id.assert_called_once()
        G.client.execute.assert_not_called()

    def test_experiment_logs_succeeded(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock(Client, autospec=True)
        G.get_experiment_id = MagicMock(return_value='some-id')
        G.experiment_details = MagicMock(return_value={'getExperimentDetails': {'status': 'succeeded'}})
        G.client.execute = MagicMock(
            return_value={
                'getArchiveExperimentLogs': {
                    'totalPages': 2,
                    'lines': [{
                        "message": "foo",
                        "timestamp": "date1"
                    }, {
                        "message": "bar",
                        "timestamp": "date2"
                    }]
                }
            })
        G.experiment_logs('test')
        G.experiment_details.assert_called_once()
        G.get_experiment_id.assert_called_once()
        G.client.execute.assert_called_once()

    def test_experiment_logs_running(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock(Client, autospec=True)
        G._init_client = MagicMock()
        G.experiment_details = MagicMock(
            return_value={'getExperimentDetails': {
                'status': 'running',
                'experimentId': 'test'
            }})
        G.client.subscribe = MagicMock(return_value=[{
            'getLiveExperimentLogs': [{
                "message": "foo",
                "timestamp": "date1"
            }, {
                "message": "bar",
                "timestamp": "date2"
            }]
        }])

        G.experiment_logs('test')
        G._init_client.assert_called_once()
        G.client.subscribe.assert_called_once()

    def test_experiment_logs_deleted(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock()
        G.get_experiment_id = MagicMock(return_value='some-id')
        G.experiment_details = MagicMock(
            return_value={'getExperimentDetails': {
                'status': 'deleted',
                'experimentId': 'test'
            }})
        with pytest.raises(ClickException) as ex_info:
            G.experiment_logs('test')
        assert "Could not find experiment" in ex_info.value.args[0]
        G.client.execute.assert_not_called()

    def test_experiment_metrics_queued(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock()
        G.get_experiment_id = MagicMock(return_value='some-id')
        G.experiment_details = MagicMock(
            return_value={'getExperimentDetails': {
                'status': 'queued',
                'experimentId': 'test'
            }})
        G.experiment_metrics('test', 'train_loss')
        G.experiment_details.assert_called_once()
        G.get_experiment_id.assert_called_once()
        G.client.execute.assert_not_called()

    def test_experiment_metrics_succeeded(self):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        G.client = MagicMock(Client, autospec=True)
        G.get_experiment_id = MagicMock(return_value='some-id')
        G.experiment_details = MagicMock(return_value={'getExperimentDetails': {'status': 'succeeded'}})

        # TODO: use when format = 'binary'
        # steps = base64.b64encode(zlib.compress(b'00000000'))
        # values = base64.b64encode(zlib.compress(b'00000000'))
        G.client.execute = MagicMock(
            return_value={'getExperimentScalarLogs': {
                'steps': "[0, 1]",
                'values': "[0.1, 0.2]",
                'nextIndex': 10
            }})
        G.experiment_metrics('test', 'train_loss')
        G.client.execute.assert_called_once()

    @pytest.mark.skip
    @httpretty.activate
    def test_check_version_compatibility(self, capsys, monkeypatch):
        G = grid.Grid(credential_path=self.creds_path, load_local_credentials=False)

        url = env.GRID_URL.replace('graphql', '')
        httpretty.register_uri(httpretty.GET, urljoin(url, 'metadata'), body=json.dumps({'version': __version__}))
        G._check_version_compatibility()

        captured = capsys.readouterr()
        assert 'pip install' not in captured.out

        httpretty.register_uri(httpretty.GET, urljoin(url, 'metadata'), body=json.dumps({'version': 'v100.1.1'}))

        try:
            G._check_version_compatibility()
        except Exception as e:
            captured = capsys.readouterr()
            assert 'pip install' in captured.out

            with pytest.raises(click.ClickException):
                raise e

        # Check that version check skips
        monkeypatch.setattr(env, 'SKIP_VERSION_CHECK', True)
        G._check_version_compatibility()

    @staticmethod
    def test_gen_ssh_config():
        assert grid.Grid._gen_ssh_config([], "", start_marker="#START", end_marker="#END") == '\n'.join([
            "#START",
            "#END",
        ])

        assert grid.Grid._gen_ssh_config([], "happy little config", start_marker="#START",
                                         end_marker="#END") == '\n'.join([
                                             "happy little config",
                                             "#START",
                                             "#END",
                                         ])

        with pytest.raises(ValueError):
            grid.Grid._gen_ssh_config([], '\n'.join(["#START"]), start_marker="#START", end_marker="#END")

        with pytest.raises(ValueError):
            grid.Grid._gen_ssh_config([], '\n'.join(["#END"]), start_marker="#START", end_marker="#END")

        with pytest.raises(ValueError):
            grid.Grid._gen_ssh_config([],
                                      '\n'.join([
                                          "#START",
                                          "#END",
                                          "#START",
                                      ]),
                                      start_marker="#START",
                                      end_marker="#END")

        with pytest.raises(ValueError):
            grid.Grid._gen_ssh_config([],
                                      '\n'.join([
                                          "#START",
                                          "#START",
                                          "#END",
                                      ]),
                                      start_marker="#START",
                                      end_marker="#END")

        with pytest.raises(ValueError):
            grid.Grid._gen_ssh_config([], '\n'.join([
                "#END",
                "#START",
            ]), start_marker="#START", end_marker="#END")

        assert grid.Grid._gen_ssh_config([
            {
                'name': 'ix1',
                'ssh_url': 'ssh://jovyan@ix1.grid.ai:233',
                'status': 'running',
            },
            {
                'name': 'ix2',
                'ssh_url': 'ssh://jovyan@ix2.grid.ai:433',
                'status': 'running',
            },
        ],
                                         '\n'.join([
                                             "happy little config",
                                             "#START",
                                             "#END",
                                         ]),
                                         start_marker="#START",
                                         end_marker="#END") == '\n'.join([
                                             "happy little config",
                                             "#START",
                                             "Host ix1",
                                             "    User jovyan",
                                             "    Hostname ix1.grid.ai",
                                             "    Port 233",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "Host ix2",
                                             "    User jovyan",
                                             "    Hostname ix2.grid.ai",
                                             "    Port 433",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "#END",
                                         ])

    @staticmethod
    def test_gen_ssh_config_idempotent():
        assert grid.Grid._gen_ssh_config([
            {
                'name': 'ix1',
                'ssh_url': 'ssh://jovyan@ix1.grid.ai:233',
                'status': 'running',
            },
            {
                'name': 'ix2',
                'ssh_url': 'ssh://jovyan@ix2.grid.ai:433',
                'status': 'running',
            },
            {
                'name': 'ix2',
                'ssh_url': 'ssh://jovyan@ix2.grid.ai:433',
                'status': 'pending',
            },
        ],
                                         '\n'.join([
                                             "happy little config",
                                             "#START",
                                             "Host ix1",
                                             "    User jovyan",
                                             "    Hostname ix1.grid.ai",
                                             "    Port 233",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "Host ix2",
                                             "    User jovyan",
                                             "    Hostname ix2.grid.ai",
                                             "    Port 433",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "#END",
                                         ]),
                                         start_marker="#START",
                                         end_marker="#END") == '\n'.join([
                                             "happy little config",
                                             "#START",
                                             "Host ix1",
                                             "    User jovyan",
                                             "    Hostname ix1.grid.ai",
                                             "    Port 233",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "Host ix2",
                                             "    User jovyan",
                                             "    Hostname ix2.grid.ai",
                                             "    Port 433",
                                             "    StrictHostKeyChecking accept-new",
                                             "    CheckHostIP no",
                                             "#END",
                                         ])



import logging
from pathlib import Path
import tempfile
from unittest.mock import MagicMock

from click.testing import CliRunner
from tests import resolvers
from tests.mock_backend import GridAIBackenedTestServer

from grid import cli
from grid.client import Grid
import grid.client as grid

logger = logging.getLogger()


def invoke(*args):
    result = CliRunner().invoke(cli.ssh_keys, args)
    output = result.stdout_bytes.decode('utf-8')
    logger.info(f"\n===== ssh-key {' '.join(args)} =====\n" + output)
    assert result.exit_code == 0
    return output


class TestSSHKey:
    def test_ssh_keys_flow(self, monkeypatch):
        # Disable this mock to test against real API server
        monkeypatch.setattr(Grid, '_init_client', MagicMock())
        monkeypatch.setattr(Grid, 'execute_gql', MagicMock())

        key_pub = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIL2cDUezQeiAzETVPCe4O1NI/U+04nYIeDHZs3uWPWOC nmiculinic@atlas"

        with tempfile.TemporaryDirectory() as tmpdirname:
            key_path_pub = Path(tmpdirname) / "id_ed25519.pub"
            key_path_pub.write_text(key_pub)

            key_id = invoke('add', "my key", str(key_path_pub))[len('Added key with id: '):].strip()
            logger.info(f"key_id {key_id}")
            invoke('list')
            invoke('authorized_keys')
            invoke('rm', key_id)
            invoke('list')

    def test_list(self, monkeypatch):
        resolver = MagicMock(return_value=[{
            "id": "8156af4e-b7fc-466f-9cbf-a3db840fff45",
            "publicKey": "AAAA GGGG ZZZZ",
            "name": "my key"
        }])

        def monkey_patch_client(x):
            x.client = GridAIBackenedTestServer()

        monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)
        monkeypatch.setattr(grid, 'gql', lambda x: x)
        monkeypatch.setattr(resolvers, 'get_public_ssh_keys', resolver)
        result = invoke("list")
        resolver.assert_called()
        assert "8156af4e-b7fc-466f-9cbf-a3db840fff45" in result
        assert "AAAA GGGG ZZZZ" in result

    @staticmethod
    def test_add_ssh_key(monkeypatch):
        resolver = MagicMock(return_value={
            "message": "ok",
            "success": True,
            "id": "8156af4e-b7fc-466f-9cbf-a3db840fff45",
        })

        def monkey_patch_client(x):
            x.client = GridAIBackenedTestServer()

        monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)
        monkeypatch.setattr(grid, 'gql', lambda x: x)
        monkeypatch.setattr(resolvers, 'add_ssh_public_key', resolver)
        key_pub = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIL2cDUezQeiAzETVPCe4O1NI/U+04nYIeDHZs3uWPWOC nmiculinic@atlas"
        with tempfile.TemporaryDirectory() as tmpdirname:
            key_path_pub = Path(tmpdirname) / "id_ed25519.pub"
            key_path_pub.write_text(key_pub)
            result = invoke("add", "my key", str(key_path_pub)).strip()
            assert result == 'Added key my key with id: 8156af4e-b7fc-466f-9cbf-a3db840fff45'

        resolver.assert_called()

    def test_remove_ssh_key(self, monkeypatch):
        resolver = MagicMock(return_value={
            "success": True,
            "message": "deleted",
        })

        def monkey_patch_client(x):
            x.client = GridAIBackenedTestServer()

        monkeypatch.setattr(grid.Grid, '_init_client', monkey_patch_client)
        monkeypatch.setattr(grid, 'gql', lambda x: x)
        monkeypatch.setattr(resolvers, 'delete_ssh_public_key', resolver)

        invoke("rm", "8156af4e-b7fc-466f-9cbf-a3db840fff45")
        _, kwargs = resolver.call_args  # skipcq:PYL-E0633
        assert kwargs == {'id': '8156af4e-b7fc-466f-9cbf-a3db840fff45'}



import click

from grid.client import Grid


@click.command()
@click.argument('experiment', type=str, nargs=1)
@click.argument('metric', type=str, nargs=1)
@click.option('--n_lines', default=10, type=int, help='Number of metrics lines (from the end) to return.')
def metrics(experiment: str, metric: str, n_lines: int) -> None:
    """Gets metric values from an experiment."""
    # Instantiate client.
    client = Grid()

    # Call client.
    client.experiment_metrics(experiment_name=experiment, metric=metric, n_lines=n_lines)



from grid.commands.git import execute_git_command


def test_execute_git_command():
    """execute_git_command() executes a git command."""
    result = execute_git_command(['version'])
    assert 'git version' in result



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid

RUNNER = CliRunner()


class TestDelete:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

        create_test_credentials()

    def test_delete_fails_without_arguments(self):
        """grid delete fails without arguments"""
        result = RUNNER.invoke(cli.delete, [])
        assert result.exit_code == 0
        assert result.stdout_bytes.decode('utf-8').startswith('Usage')

    def test_delete_deletes_runs_and_experiments(self):
        """grid delete fails without arguments"""
        result = RUNNER.invoke(cli.delete, ["run", 'test-run'], input='y\n')
        assert result.exit_code == 0
        assert not result.exception
        assert 'Are you sure you want to do this?' in result.output
        assert 'deleted successfully' in result.output

        result = RUNNER.invoke(cli.delete, ["experiment", 'test-run-exp0'], input='y\n')
        assert result.exit_code == 0
        assert not result.exception
        assert 'Are you sure you want to do this?' in result.output
        assert 'deleted successfully' in result.output

    def test_delete_deletes_warning_prevents_delete(self):
        """grid delete warning prevents a deletion"""
        result = RUNNER.invoke(cli.delete, ["run", 'test-run'], input='N\n')
        assert result.exit_code == 1
        assert 'Are you sure you want to do this?' in result.output

    def test_delete_run_fails(self, monkeypatch):
        """grid delete run fails with exception"""
        def delete_run(*args, **kwargs):
            return {"success": False}

        monkeypatch.setattr(resolvers, 'delete_run', delete_run)
        result = RUNNER.invoke(cli.delete, ["run", 'test-run'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output

    def test_delete_run_fails_with_exception(self, monkeypatch):
        """grid delete run fails with exception"""
        def delete_run(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'delete_run', delete_run)
        result = RUNNER.invoke(cli.delete, ["run", 'test-run'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output

    def test_delete_experiment_fails(self, monkeypatch):
        """grid delete experiment fails"""
        def delete_experiment(*args, **kwargs):
            return {"success": False}

        monkeypatch.setattr(resolvers, 'delete_experiment', delete_experiment)
        result = RUNNER.invoke(cli.delete, ["experiment", 'test-run-exp0'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output

    def test_delete_experiment_fails_with_exception(self, monkeypatch):
        """grid delete experiment fails with exception"""
        def delete_experiment(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'delete_experiment', delete_experiment)
        result = RUNNER.invoke(cli.delete, ["experiment", 'test-run-exp0'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output

    def test_delete_experiment_fails_when_experiment_not_found(self, monkeypatch):
        def mp_get_experiment_id(*args, **kwargs):
            return {"success": False}

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        result = RUNNER.invoke(cli.delete, ["experiment", 'test-run-exp0'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output
        assert 'Could not find experiment' in result.output

    def test_delete_experiment_fails_when_experiment_not_found_exception(self, monkeypatch):
        def mp_get_experiment_id(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'get_experiment_id', mp_get_experiment_id)
        result = RUNNER.invoke(cli.delete, ["experiment", 'test-run-exp0'], input='y\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Delete failed' in result.output
        assert 'Could not find experiment' in result.output



from typing import Dict, List, Optional

from grid.core.artifact import Artifact
from grid.core.base import GridObject


class Experiment(GridObject):
    """
    Grid experiment object. This object contains all the properties that a
    given artifact should have. This also encapsulates special methods
    for interactive with experiment properties.

    Parameters
    ----------
    name: str
        Experiment name
    """
    def __init__(self, name: str):
        self._data: Optional[Dict[str, str]] = {}
        self.name = name
        self.identifier = None
        super().__init__()

    def refresh(self) -> None:
        """
        Updates object metadata. This makes a query to Grid to fetch the
        object's latest data.
        """
        self.identifier = self.client.get_experiment_id(self.name)

        query = """
        query GetExperimentDetails ($experimentId: ID!) {
            getExperimentDetails(experimentId: $experimentId) {
                name
                githubId
                desiredState
                commitSha
                entrypoint
                invocationCommands
                createdAt
                startedRunningAt
                finishedAt
                parameters {
                    name
                    value
                }
            }
        }
        """
        result = self.client.execute_gql(query, experimentId=self.identifier)
        self._data = result["getExperimentDetails"]
        self._update_meta()

    @property
    def status(self):
        query = """
        query GetExperimentDetails ($experimentId: ID!) {
            getExperimentDetails(experimentId: $experimentId) {
                status
            }
        }
        """
        result = self.client.execute_gql(query, experimentId=self.identifier)
        return result["getExperimentDetails"]["status"]

    @property
    def artifacts(self) -> List[Dict[str, str]]:
        """Fetches artifacts from a given experiments. Artifacts are"""
        query = """
        query (
            $experimentId: ID!
        ) {
            getArtifacts(experimentId: $experimentId) {
                signedUrl
                downloadToPath
                downloadToFilename
            }
        }
        """
        result = self.client.execute_gql(query, experimentId=self.identifier)
        return [Artifact(*a.values()) for a in result.get("getArtifacts")]



from typing import List

import click

import grid.globals as env


@click.group()
def view():
    pass


@view.command()
@click.argument('experiment_id', type=str, required=True, nargs=1)
@click.argument('page', type=str, nargs=1, required=False)
def experiment(experiment_id: List[str], page: str) -> None:
    """Grid view shows we web UI page for your runs and experiments."""
    # Fetch URL from globals.
    url = env.GRID_URL.replace('/graphql', '#')

    # Figure out which object is requested
    # so we can construct path.
    base_path = 'view'
    qualifier_path = 'experiment'

    # Combine all strings into a single URL.
    if page:
        launch_url = '/'.join([url, base_path, qualifier_path, experiment_id, page])
    else:
        launch_url = '/'.join([url, base_path, qualifier_path, experiment_id])

    # Open browser.
    click.echo()
    click.echo(f'Opening URL: {launch_url}')
    click.echo()

    click.launch(launch_url)


@view.command()
@click.argument('run_name', type=str, nargs=1)
@click.argument('page', type=str, nargs=1, required=False)
def run(run_name: str, page: str) -> None:
    """Grid view shows we web UI page for your runs and experiments."""
    # Fetch URL from globals.
    url = env.GRID_URL.replace('/graphql', '#')

    # Figure out which object is requested
    # so we can construct path.
    base_path = 'view'
    qualifier_path = 'run'

    # Combine all strings into a single URL.
    if page:
        launch_url = '/'.join([url, base_path, qualifier_path, run_name, page])
    else:
        launch_url = '/'.join([url, base_path, qualifier_path, run_name])

    # Open browser.
    click.echo()
    click.echo(f'Opening URL: {launch_url}')
    click.echo()

    click.launch(launch_url)



from pathlib import Path

import click
from rich.console import Console
from rich.table import Table

from grid.client import Grid


@click.group()
def ssh_keys() -> None:
    """Manage SSH keys in Grid"""


@ssh_keys.command(help="Register a new SSH public key and give it name.")
@click.argument('name', type=str, nargs=1, required=True)
@click.argument('key', type=click.Path(exists=True), nargs=1, required=True)
def add(name: str, key: str):
    key = Path(key).read_text('utf-8').strip()
    result = Grid().add_ssh_public_key(key=key, name=name)

    if result['addSSHPublicKey'].get('success'):
        click.echo(f"Added key {name} with id: {result['addSSHPublicKey']['id']}")

    else:
        raise click.ClickException(f"Failed to add key: {key}")


@ssh_keys.command(help="list currently registered SSH public keys", name="list")
@click.option('--limit', 'limit', type=int, required=False, default=100, help='maximum number of public keys to fetch')
def ls(limit: int):
    table = Table(show_header=True, header_style="bold green")
    table.add_column('id', style='dim')
    table.add_column('name', style='dim')
    table.add_column('public_key', style='dim')

    client = Grid()
    for row in client.list_public_ssh_keys(limit):
        table.add_row(row['id'], row['name'], row['publicKey'])
    Console().print(table)


@ssh_keys.command(name="authorized_keys",
                  hidden=True,
                  help="list all registered SSH public keys in authorized_keys format")
@click.option('--limit', 'limit', type=int, required=False, default=100, help='maximum number of public keys to fetch')
def authorized_keys(limit: int):
    for row in Grid().list_public_ssh_keys(limit):
        click.echo(f"#ID {row['id'].strip()}")
        click.echo(f"#Name {row['name'].strip()}")
        click.echo(row['publicKey'].strip())


# This command is used within ssh server AuthorizedKeysCommand section
# and it is not intended to be used by the end consumer, thus it's hidden
@ssh_keys.command(help="remote registered SSH public key")
@click.argument('key_id', type=str, nargs=1)
def rm(key_id: str):
    Grid().delete_ssh_public_key(key_id=key_id)



import pprint
from unittest.mock import patch

import click

from grid.tracking import Grid
from grid.tracking import Segment
from grid.tracking import TrackingEvents


class TestTracking:
    def test_env_sensitive(self):
        for key, is_sensitive in {
                "HOME": False,
                "PATH": False,
                "AWS_ACCESS_KEY_ID": True,
                "AWS_SECRET_ACCESS_KEY": True,
        }.items():
            assert is_sensitive == Segment._is_env_sensitive(key)

    def test_segment(self):
        def check_context(ctx):
            assert 'version' in ctx
            assert 'cwd' in ctx
            assert 'env' != ctx['env']
            assert 'json' in ctx

        with patch('analytics.Client', autospec=True):
            s = Segment()
            s.client.identify.assert_called_once()
            args, kwargs = s.client.identify.call_args
            print("Identify called", args, pprint.pformat(kwargs))
            check_context(kwargs['context'])

            s.send_event(TrackingEvents.CLI_FINISHED, {'s': 'b'})
            s.client.track.assert_called_once()
            args, kwargs = s.client.track.call_args
            print("CLI Finished called", args, pprint.pformat(kwargs))
            check_context(kwargs['context'])
            assert TrackingEvents.CLI_FINISHED.value == kwargs['event']
            assert {'s': 'b', 'json': '{"s": "b"}'} == kwargs['properties']

            error_msg = ""
            try:
                error_msg = "I'm a super cool error message"
                raise click.ClickException(error_msg)
            except Exception as e:
                s.report_exception(e)
            assert 2 == s.client.track.call_count

            args, kwargs = s.client.track.call_args
            print("exception caught", args, pprint.pformat(kwargs))
            check_context(kwargs['context'])
            props = kwargs['properties']
            assert props['error'] == 'ClickException'
            assert props['message'] == error_msg
            assert len(props['stacktrace']) > 0

            s.flush()
            s.client.flush.assert_called_once()

            class MockException:
                message = 'test'

            s.report_exception(MockException)

            _, kwargs = s.client.track.call_args
            props = kwargs['properties']
            assert props['message'] == 'test'

    def test_handles_grid_client_exception(self, monkeypatch):
        def mp_init(*args, **kwargs):
            raise click.ClickException('test failure')

        # If the exception is not handled, then this would fail.
        monkeypatch.setattr(Grid, '__init__', mp_init)
        Segment()



import click
from gql.transport.exceptions import TransportQueryError
import pytest
import tests.resolvers as resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

import grid.client as grid
import grid.observables.run as run_observable


def mp_execute(*args, **kwargs):
    raise Exception("{'message':'not found'}")


class TestRunObservable:
    @classmethod
    def setup_class(cls):
        create_test_credentials()

        grid.Grid._init_client = monkey_patch_client

        run_observable.gql = lambda x: x

        cls.grid = grid.Grid(load_local_credentials=False)
        cls.grid._init_client()

    def test_get(self):
        run_observable.env.DEBUG = True
        obsevable = run_observable.Run(client=self.grid.client)
        result = obsevable.get()
        assert len(result) > 0

    def test_get_handles_transport_error(self, monkeypatch):
        def mp_execute_transport(*args, **kwargs):
            raise TransportQueryError('test')

        monkeypatch.setattr(self.grid.client, 'execute', mp_execute_transport)
        obsevable = run_observable.Run(client=self.grid.client)
        with pytest.raises(click.ClickException):
            obsevable.get()

    def test_get_with_id(self, capsys):
        run_observable.env.DEBUG = True
        obsevable = run_observable.Run(client=self.grid.client, identifier='test-run')
        result = obsevable.get()
        assert len(result) > 0

        # captured = capsys.readouterr()
        # assert 'nExperiments' in captured.out

    def test_get_history(self, capsys):
        run_observable.env.DEBUG = True
        obsevable = run_observable.Run(client=self.grid.client, identifier='test-run')
        result = obsevable.get_history()

        assert len(result) > 0
        assert 'createdAt' in result['getRuns'][0]

        captured = capsys.readouterr()
        assert 'Created At' in captured.out

    def test_get_shows_empty_table(self, capsys, monkeypatch):
        def mp_get_runs(*args, **kwargs):
            return [{
                'name': 'test-run-0',
                'createdAt': resolvers.now,
                'experiments': [{
                    'experimentId': 'test-run-exp0'
                }],
                'nExperiments': 2,
                'nFailed': 2,
                'nCancelled': 0,
                'nRunning': 0,
                'nCompleted': 0,
                'nQueued': 0,
                'nPending': 0,
                'projectId': 'test/project'
            }]

        monkeypatch.setattr(resolvers, 'get_runs', mp_get_runs)
        obsevable = run_observable.Run(client=self.grid.client)
        obsevable.get()
        captured = capsys.readouterr()
        assert 'Run' in captured.out

    def test_queued_and_pending_are_shown_in_same_column(self, capsys, monkeypatch):
        def mp_get_runs(*args, **kwargs):
            return [{
                'name': 'test-run-0',
                'createdAt': resolvers.now,
                'experiments': [{
                    'experimentId': 'test-run-exp0'
                }],
                'nExperiments': 2,
                'nFailed': 0,
                'nCancelled': 0,
                'nRunning': 0,
                'nCompleted': 0,
                'nQueued': 1,
                'nPending': 1,
                'projectId': 'test/project'
            }]

        monkeypatch.setattr(resolvers, 'get_runs', mp_get_runs)
        obsevable = run_observable.Run(client=self.grid.client)
        obsevable.get()
        captured = capsys.readouterr()
        assert 'Run' in captured.out

        # count that nQueued and nPending are added together,
        # which will make this show-up twice
        assert captured.out.count('2') == 2

    def test_get_shows_empty_history_table(self, capsys, monkeypatch):
        def mp_get_runs(*args, **kwargs):
            return []

        monkeypatch.setattr(resolvers, 'get_runs', mp_get_runs)
        obsevable = run_observable.Run(client=self.grid.client)
        obsevable.get_history()
        captured = capsys.readouterr()
        assert 'No History' in captured.out



import click

from grid.client import Grid


@click.group()
def delete() -> None:
    pass


def doublecheck(item: str):
    warning_str = click.style('WARNING!', fg='red')
    message = f"""

    {warning_str}

    Your are about to delete the {item}.
    This will delete all the associated artifacts, logs, and metadata.

    Are you sure you want to do this?

   """
    click.confirm(message, abort=True)


@delete.command()
@click.argument('experiment_ids', type=str, required=True, nargs=-1)
def experiment(experiment_ids: [str]):
    doublecheck(experiment_ids)
    client = Grid()
    for experiment in experiment_ids:
        client.delete(experiment_name=experiment)


@delete.command()
@click.argument('run_ids', type=str, required=True, nargs=-1)
def run(run_ids: [str]):
    doublecheck(run_ids)
    client = Grid()
    for run in run_ids:
        client.delete(run_name=run)



from typing import Optional

import click

from grid.client import Grid


@click.command()
@click.argument('experiment', type=str, nargs=1)
@click.option('--n_lines', default=10, type=int, help='Number of log lines (from the end) to return.')
@click.option('--page', type=int, help='Which page to fetch from archived logs')
@click.option('--max_lines', type=int, help='Maximum number of lines to print in terminal')
@click.option('--use_pager',
              is_flag=True,
              default=False,
              help='If the log output should be displayed as a pager scrollable')
def logs(experiment: str, n_lines: int, page: Optional[int], max_lines: Optional[int], use_pager: bool) -> None:
    """Gets logs from an experiment."""
    # Instantiate client.
    client = Grid()

    # Call client.
    client.experiment_logs(experiment_name=experiment,
                           n_lines=n_lines,
                           page=page,
                           max_lines=max_lines,
                           use_pager=use_pager)



import random
import subprocess
import sys
from typing import Dict, Optional

import click
from coolname import generate_slug
import yaml

from grid.cli.grid_credentials import print_creds_table
from grid.cli.grid_train import _check_run_name_is_valid, _resolve_instance_type_nickname, _get_instance_types
from grid.cli.utilities import read_config
from grid.cli.utilities import validate_config
from grid.cli.utilities import validate_disk_size_callback
from grid.client import Grid
import grid.globals as env
from grid.tracking import Segment
from grid.tracking import TrackingEvents
from grid.types import ObservableType
from grid.utilities import check_description_isnt_too_long


def _print_submission_message(instance_type: str, provider: str, credential_id: str, name: str, datastore_name: str,
                              datastore_version: str, datastore_mount_dir: str):
    """Prints train submission message to the client."""
    message = f"""
        Interactive node created!

        `grid status` to list all runs and interactive nodes.
        `grid status {name}` to see the status fot this interactive node.

        ----------------------
        Submission summary
        ----------------------
        name:                    {name}
        instance_type:           {instance_type}
        cloud_provider:          {provider}
        cloud_credentials:       {credential_id}
        datastore_name:          {datastore_name}
        datastore_version:       {datastore_version}
        datastore_mount_dir:     {datastore_mount_dir}
        """

    click.echo(message)


def _generate_default_interactive_config(client: Grid,
                                         credential_id: str,
                                         instance_type: str,
                                         datastore_name: Optional[str] = None,
                                         datastore_version: Optional[str] = None,
                                         datastore_mount_dir: Optional[str] = None,
                                         disk_size: Optional[int] = None):
    """
    Generates a new default config file for user if user hasn't
    submitted one.
    """
    _default_credential = None
    creds = client.get_credentials()['getUserCredentials']
    if credential_id:
        if creds:
            #  Check if the credential passed is can
            #  be used by user. We'll raise an exception
            #  if user can't use such credential.
            for cred in creds:
                if cred['credentialId'] == credential_id:
                    _default_credential = cred
                    break

        if not _default_credential:
            raise click.ClickException(f'Credential ID {credential_id} does not exist. ' +
                                       'Use grid credentials to see available credential IDs.')
    else:
        _url = client.url.replace('/graphql', '')
        no_credentials_message = f"""
        No cloud credentials available! Visit {_url}/#/settings to
        add new cloud credentials.
        """
        if not creds:
            raise click.ClickException(no_credentials_message)

        if len(creds) == 1:
            cred = creds[0]
            _default_credential = cred

        elif len(creds) > 1:
            for credential in creds:
                if credential['defaultCredential']:
                    click.echo(f"Using default cloud credentials {credential['credentialId']} "
                               f"to run on {credential['provider'].upper()}.")
                    _default_credential = credential
                    break

        #  If no credentials are available, raise an exception.
        if not _default_credential:
            m = """
            Detected multiple credentials. Which would you like to use?

                grid interactive --grid_credential [CREDENTIAL_ID]

            """
            print_creds_table(creds)
            raise click.ClickException(m)

    #  Relevant defaults for user.
    #  TODO: Pull from user profile when we have
    #  that information collected.
    defaults = {'region': 'us-east-1'}

    _grid_config = {
        'compute': {
            'provider': {
                'vendor': _default_credential['provider'],
                'credentials': _default_credential['credentialId'],
                'region': defaults['region']
            },
            'train': {
                'instance': instance_type,
                'disk_size': disk_size,
                #  These parameters are required for the
                #  training config. The max_nodes need to be a value
                #  other than 0, otherwise subsequent train operations
                #  will not work.
                'nodes': 0,  # Start with 0 compute nodes.
                'max_nodes': 10,  # Scale at most to 100 nodes.
                'gpus': 0,  # Required, but ignored.
                'datastore_name': datastore_name,
                'datastore_version': datastore_version,
                'datastore_mount_dir': datastore_mount_dir,
                'scale_down_seconds': 30 * 60,  # How many seconds before scaling down?
            }
        }
    }

    #  Print debug message if in debug mode.
    if env.DEBUG:
        click.echo('Grid Config used:')
        click.echo(_grid_config)

    return _grid_config


@click.group(invoke_without_command=True)
@click.pass_context
def interactive(ctx) -> None:
    """Displays interactive nodes for given project."""
    client = Grid()

    if ctx.invoked_subcommand is None:
        # Get the status of the interactive observables.
        kind = ObservableType.INTERACTIVE
        client.status(kind=kind, follow=False)


@interactive.command()
@click.option('--grid_config',
              '--g_config',
              'config',
              type=click.File('r'),
              required=False,
              callback=read_config,
              help='Path to Grid config YML')
@click.option('--grid_name',
              '--g_name',
              'name',
              type=str,
              required=False,
              help='Name for this run',
              callback=_check_run_name_is_valid)
@click.option('--grid_description',
              '--g_description',
              'description',
              type=str,
              required=False,
              help='Interactive node description; useful for note-keeping',
              callback=check_description_isnt_too_long)
@click.option('--grid_credential', '--g_credential', 'credential', type=str, required=False, help='Cloud to run on')
@click.option('--grid_instance_type',
              '--g_instance_type',
              'instance_type',
              type=str,
              default='2_m60_8gb',
              callback=_resolve_instance_type_nickname,
              help='Instance type to start training session in',
              autocompletion=_get_instance_types)
@click.option('--grid_disk_size',
              '--g_disk_size',
              'disk_size',
              type=int,
              required=False,
              default=200,
              callback=validate_disk_size_callback,
              help='The disk size in GB to allocate to interactive node')
@click.option('--grid_datastore_name',
              '--g_datastore_name',
              'datastore_name',
              type=str,
              required=False,
              help='Datastore name to be mounted in interactive node')
@click.option('--grid_datastore_version',
              '--g_datastore_version',
              'datastore_version',
              type=str,
              required=False,
              help='Datastore version to be mounted in interactive node')
@click.option('--grid_datastore_mount_dir',
              '--g_datastore_mount_dir',
              'datastore_mount_dir',
              type=str,
              required=False,
              default='/home/jovyan/datastore',
              help='Absolute path to mount Datastore in interactive node. (default to ~/datastore)')
def create(config: Optional[Dict], name: Optional[str], description: Optional[str], credential: Optional[str],
           instance_type: str, datastore_name: Optional[str], datastore_version: Optional[str],
           datastore_mount_dir: Optional[str], disk_size: Optional[int]) -> None:
    """Creates an interactive node."""
    client = Grid()

    tracker = Segment()
    tracker.send_event(event=TrackingEvents.INTERACTIVE_NODE_CREATED, properties={'user_input': " ".join(sys.argv)})

    # make a fun random name when user does not pass in a name
    if name is None:
        name = f'{generate_slug(2)}-{random.randint(0, 1000)}'
        click.echo(f'No --grid_name passed, naming your interactive node: {name}')

    #  If the user has not passed a grid config file,
    #  then generate one with a set of default options.
    #  We'll add a default instance and the user's
    #  default credentials.
    if not config:
        _grid_config = _generate_default_interactive_config(client=client,
                                                            datastore_name=datastore_name,
                                                            datastore_version=datastore_version,
                                                            datastore_mount_dir=datastore_mount_dir,
                                                            credential_id=credential,
                                                            instance_type=instance_type,
                                                            disk_size=disk_size)

    else:
        _grid_config = config

    validate_config(cfg=_grid_config)

    client.validate_datastore_version(grid_config=_grid_config)
    datastore_version = _grid_config["compute"]["train"].get("datastore_version")

    _grid_config_yaml = yaml.dump(_grid_config)

    #  Send to client.
    client.create_interactive_node(config=_grid_config_yaml, name=name, description=description)

    #  Shows a friendly submission message to the
    #  user.
    _print_submission_message(instance_type=_grid_config['compute']['train']['instance'],
                              provider=_grid_config['compute']['provider']['vendor'],
                              credential_id=_grid_config['compute']['provider']['credentials'],
                              datastore_name=datastore_name,
                              datastore_version=datastore_version,
                              datastore_mount_dir=datastore_mount_dir,
                              name=name)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def pause(interactive_node: str) -> None:
    """Pauses an interactive node."""
    client = Grid()
    client.pause_interactive_node(interactive_node_name=interactive_node)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def resume(interactive_node: str) -> None:
    """Resumes an interactive node."""
    client = Grid()
    client.resume_interactive_node(interactive_node_name=interactive_node)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def delete(interactive_node: str) -> None:
    """Deletes an interactive node."""
    client = Grid()
    client.delete_interactive_node(interactive_node_name=interactive_node)


def _update_ssh_config_and_check_ixsession_status(ctx, param, value: str) -> int:
    """
    This updates the SSH config for interactive nodes and
    also checks if those nodes can be interacted with SSH
    before attempting an operation. This prevents SSHing into
    nodes that are not un a running state, e.g. paused or pending.

    This manages a section within user's ssh config file
    for all interactive nodes shh config details

    Afterwards you can use systems ssh & related utilities
    (sshfs, rsync, ansible, whatever) with interactive nodes
    directly

    The default file is ~/.ssh/config and  can be changed via
    envvar GRID_SSH_CONFIG

    Returns
    --------
    value: str
        Unmodified value if valid
    """
    client = Grid()
    nodes = client.sync_ssh_config()

    click.echo(f"Sync config for {len(nodes)} interactive nodes")

    target_ixsession = None
    for node in nodes:
        if node["name"] == value:
            target_ixsession = node

    # Check if interactive session exists at all
    if not target_ixsession:
        session_names = [n["name"] for n in nodes]
        raise click.BadArgumentUsage(f"Interactive session {value} does not exist. "
                                     f"Available Interactive Sessions are: {', '.join(session_names)}")

    # Check if the node is in 'running' status
    if target_ixsession["status"] != "running":
        running_ixsessions = [n["name"] for n in nodes if n["status"] == "running"]
        raise click.BadArgumentUsage(f"Interactive session {value} is not ready. "
                                     f"Sessions that are ready to SSH are: {', '.join(running_ixsessions)}")

    return value


@interactive.command()
@click.argument('node_name', type=str, callback=_update_ssh_config_and_check_ixsession_status)
def ssh(node_name):
    """
    SSH to interactive nodes.

    Usage: grid interactive ssh <node name>

    This command sync your ~/.ssh/config entries which are known to grid
    and executes ssh <node name>
    If you'd like the full power of ssh, you can use any ssh client and
    do `ssh <node_name>`. This command is stripped down version of it.

    See Also:
        grid interactive sync-ssh-config --help
    \f
    """
    client = Grid()

    client.sync_ssh_config()
    subprocess.check_call(['ssh', node_name])


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1, callback=_update_ssh_config_and_check_ixsession_status)
@click.argument('mount_dir', type=str, nargs=1)
def mount(interactive_node, mount_dir):
    """
    Mount interactive node directory to local.

    To mount a filesystem use:
    ixNode:[dir] mountpoint

    Examples:
        # Mounts the home directory on the interactive node in dir data
        grid interactive mount bluberry-122 ./data

        # mounts ~/data directory on the interactive node to ./data
        grid interactive mount bluberry-122:~/data ./data

   To unmount it:
      fusermount3 -u mountpoint   # Linux
      umount mountpoint           # OS X, FreeBSD

    Under the hood this is just passing data to sshfs after syncing grid's interactive,
    i.e. this command is dumbed down sshfs

    See Also:
        grid interactive sync-ssh-config --help
    \f
    """
    if ':' not in interactive_node:
        interactive_node += ":/home/jovyan"

    client = Grid()
    client.sync_ssh_config()
    subprocess.check_call(['sshfs', interactive_node, mount_dir])



from grid.cli.grid_artifacts import artifacts
from grid.cli.grid_cancel import cancel
from grid.cli.grid_credentials import credentials
from grid.cli.grid_datastores import datastores
from grid.cli.grid_delete import delete
from grid.cli.grid_history import history
from grid.cli.grid_interactive import interactive
from grid.cli.grid_login import login
from grid.cli.grid_logs import logs
from grid.cli.grid_metrics import metrics
# SLURM
from grid.cli.grid_slurm import slurm
from grid.cli.grid_ssh_keys import ssh_keys
from grid.cli.grid_status import status
from grid.cli.grid_train import train
from grid.cli.grid_view import view

__all__ = [
    'view',
    'status',
    'login',
    'train',
    'cancel',
    'credentials',
    'history',
    'logs',
    'metrics',
    'interactive',
    'artifacts',
    'delete',
    'datastores',
    'slurm',
    'ssh_keys',
]



from tests.utilities import monkey_patch_client

import grid.client as grid
from grid.core import Experiment
from grid.core import Run
from grid.core.task import Task


class TestRun:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

    def test_run_refresh(self):
        R = Run("test")
        R.refresh()

        assert R.name
        assert R.run_id

    def test_experiment_artifacts(self):
        R = Run("test")

        assert len(R.experiments) > 0
        for experiment in R.experiments:
            assert isinstance(experiment, Experiment)

    def test_tasks(self):
        R = Run("test")
        assert len(R.tasks) > 0
        assert all(isinstance(t, Task) for t in R.tasks)



from tests.utilities import monkey_patch_client

import grid.client as grid
from grid.core import Artifact
from grid.core import Experiment


class TestExperiment:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

    def test_experiment_refresh(self):
        E = Experiment("test")
        E.refresh()

        assert E.name
        assert E.status

    def test_experiment_artifacts(self):
        E = Experiment("test")
        E.refresh()

        assert len(E.artifacts) > 0
        for artifact in E.artifacts:
            assert isinstance(artifact, Artifact)



from typing import List

from grid.core import Experiment
from grid.core.base import GridObject
from grid.core.task import Task


class Run(GridObject):
    """
    Run object in Grid. Runs are collections of Experiment objects.

    Parameters
    ----------
    identifier: str
        Run name (not Run ID)
    """
    def __init__(self, identifier: str):
        self.identifier = identifier
        super().__init__()

    def refresh(self) -> None:
        """
        Updates object metadata. This makes a query to Grid to fetch the
        object's latest data.
        """
        query = """
        query GetRunDetails ($runName: ID!) {
            getRuns(runName: $runName) {
                runId
                name
                description
                entrypoint
                createdAt
                startedRunningAt
                finishedAt
                clusterId
                nExperiments
                nRunning
                nFailed
                nCompleted
                nCancelled
                nQueued
                nPending
                invocationCommand
                projectId
                config {
                    compute
                }
            }
        }
        """
        result = self.client.execute_gql(query, runName=self.identifier)
        self._data = result["getRuns"][0]
        self._update_meta()

    @property
    def experiments(self) -> List[Experiment]:
        """
        List of experiments for the Run.

        Returns
        -------
        experiments: List[Experiment]
            List of Experiment instances.
        """
        query = """
        query (
                $runName: ID
            ) {
                getExperiments (runName: $runName) {
                    experimentId
                    name
                    commitSha
                    entrypoint
                    invocationCommands
                    createdAt
                    finishedAt
                    startedRunningAt
                    desiredState
                }
            }
        """
        result = self.client.execute_gql(query, runName=self.identifier)

        # Skips the need for the Experiment object to reload
        # data from the backend API.
        experiments = []
        for experiment_data in result.get("getExperiments"):
            E = Experiment(experiment_data.pop("experimentId"))
            E._data = experiment_data
            E._update_meta()
            experiments.append(E)

        return experiments

    @property
    def tasks(self) -> List[Task]:
        query = """
        query (
            $runName: ID!
        ) {
            getRunTaskStatus (
                runName: $runName
            ) {
                dependencies {
                    taskId
                    status
                    taskType
                    message
                    error
                }
            }
        }
        """
        result = self.client.execute_gql(query, runName=self.identifier)
        tasks = result["getRunTaskStatus"]["dependencies"]

        task_objects = []
        for task in tasks:
            dict_task = {self._camel_case_to_snake_case(k): v for k, v in task.items()}
            task_objects.append(Task(**dict_task))

        return task_objects



import subprocess
from typing import List


def execute_git_command(args: List[str]) -> str:
    """
    Executes a git command. This is expected to return a
    single string back.

    Returns
    -------
    output: str
        String combining stdout and stderr.
    """
    process = subprocess.run(['git'] + args,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True,
                             check=False)

    output = process.stdout.strip() + process.stderr.strip()
    return output



from pathlib import Path
from typing import List, Optional

import click
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TimeRemainingColumn

from grid.core import Artifact
from grid.core import Experiment
from grid.core import Run
from grid.downloader import DownloadableObject
from grid.downloader import Downloader
from grid.utilities import is_experiment


def _download_artifacts(artifacts: List[Artifact], download_dir: str) -> None:
    """
    Downloads a set of artifacts from Grid.

    Parameters
    ----------
    artifacts: List[Artifact]
        List of Artifact objects.

    download_dir: str
        Download directory for artifacts
    """
    # Create host directory.
    Downloader.create_dir_tree(dest_dir=download_dir)

    # Create downloadable objects.
    files_to_download = []
    for artifact in artifacts:
        files_to_download.append(
            DownloadableObject(url=artifact.url, download_path=artifact.path, filename=artifact.filename))

    # Start download if there are any files to download.
    if files_to_download:
        D = Downloader(downloadable_objects=files_to_download, base_dir=download_dir)
        D.download()


@click.command()
@click.option("--download_dir",
              type=click.Path(exists=False, file_okay=False, dir_okay=True),
              required=False,
              default="./grid_artifacts",
              help="Download directory that will host all artifact files")
@click.argument("runs_or_experiments", type=str, required=True, nargs=-1)
def artifacts(runs_or_experiments: List[str], download_dir: Optional[str] = None) -> None:
    """Downloads artifacts for a given run or experiments."""
    progress = Progress(
        SpinnerColumn(),
        "[progress.description]{task.fields[file]}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        TimeRemainingColumn(),
    )
    task = progress.add_task("downloading", total=len(runs_or_experiments), start=True, file="")

    click.echo(f"Downloading artifacts → {runs_or_experiments}")
    with progress:
        for element in runs_or_experiments:
            progress.tasks[task].fields["file"] = element
            if is_experiment(element):
                experiment = Experiment(name=element)
                experiment.refresh()
                _download_artifacts(experiment.artifacts, download_dir=download_dir)
                progress.update(task, advance=1)
            else:
                # Runs need their own host directory to make
                # it different than experiments.
                host_path = Path(download_dir) / Path(element)
                run = Run(identifier=element)

                for experiment in run.experiments:
                    progress.tasks[task].fields["file"] = element
                    experiment.refresh()
                    _download_artifacts(experiment.artifacts, download_dir=host_path)

                progress.update(task, advance=1)



import click
import pytest
import tests.resolvers as resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

import grid.client as grid
import grid.observables.base as base_observables


def mp_execute(*args, **kwargs):
    raise Exception("{'message':'not found'}")


def test_style_status():
    value = 'test'
    result = base_observables.style_status(format_string=value, status='running')
    assert '\x1b[33m' in result
    assert value in result

    result = base_observables.style_status(format_string=value, status='failed')
    assert '\x1b[31m' in result
    assert value in result

    result = base_observables.style_status(format_string=value, status='finished')
    assert '\x1b[32m' in result
    assert value in result

    result = base_observables.style_status(format_string=value, status='cancelled')
    assert '\x1b[37m' in result
    assert value in result


class TestBaseObservable:
    @classmethod
    def setup_class(cls):
        create_test_credentials()

        grid.Grid._init_client = monkey_patch_client

        base_observables.gql = lambda x: x

        cls.grid = grid.Grid(load_local_credentials=False)
        cls.grid._init_client()

        cls.observable = base_observables.BaseObservable(client=cls.grid.client)

    def test_get_task_run_dependencies(self):
        results = self.observable._get_task_run_dependencies(run_name='test')
        assert len(results) > 0

    def test_get_task_run_dependencies_raise_exception(self, monkeypatch):
        monkeypatch.setattr(self.grid.client, 'execute', mp_execute)
        observable = base_observables.BaseObservable(client=self.grid.client)
        with pytest.raises(click.ClickException):
            observable._get_task_run_dependencies(run_name='test-run')

    def test_get_task_run_status(self, capsys):
        base_observables.env.SHOW_PROCESS_STATUS_DETAILS = False
        base_observables.env.DEBUG = True
        self.observable._get_task_run_status(run_name='test')

        captured = capsys.readouterr()
        assert 'grid status' in captured.out

        base_observables.env.SHOW_PROCESS_STATUS_DETAILS = True
        base_observables.env.DEBUG = True
        self.observable._get_task_run_status(run_name='test')

        captured = capsys.readouterr()
        assert 'Failed error details' in captured.out

    def test_get_task_run_status_with_non_user_facing_tasks(self, monkeypatch):
        base_observables.env.SHOW_PROCESS_STATUS_DETAILS = False
        base_observables.env.DEBUG = True

        def mp_get_run_task_status(*args, **kwargs):
            return {
                'success':
                True,
                'runId':
                'test',
                'name':
                '',
                'status':
                'failed',
                'message':
                '',
                'dependencies': [{
                    'taskId': 'test-task-0',
                    'status': 'failed',
                    'taskType': 'grid.core.repository_builder.RepositoryBuilder',
                    'message': 'Failed message',
                    'error': 'Failed error details'
                }, {
                    'taskId': 'test-task-1',
                    'status': 'running',
                    'taskType': 'grid.core.user.ReconcileUser',
                    'message': '',
                    'error': ''
                }, {
                    'taskId': 'test-task-2',
                    'status': 'running',
                    'taskType': 'grid.core.clusters.global_user_cluster.ReconcileCluster',
                    'message': '',
                    'error': ''
                }]
            }

        monkeypatch.setattr(resolvers, 'get_run_task_status', mp_get_run_task_status)
        self.observable._get_task_run_status(run_name='test')



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid

RUNNER = CliRunner()


class TestCancel:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client

        grid.gql = lambda x: x

        create_test_credentials()

        cls.cluster_alias = 'cluster_alias'

    def test_slurm_logs_succeeds(self):
        """grid slurm logs succeeds"""
        result = RUNNER.invoke(cli.slurm, ["logs"])
        assert result.exit_code == 0
        assert not result.exception

    def test_slurm_cancel_succeeds(self):
        """grid slurm cancel succeeds"""
        result = RUNNER.invoke(cli.slurm, ["cancel"])
        assert result.exit_code == 0
        assert not result.exception

    def test_slurm_status_fails(self, monkeypatch):
        """grid slurm status fails"""
        def get_slurm_status(*args, **kwargs):
            return {
                'success': False,
            }

        monkeypatch.setattr(resolvers, 'get_slurm_status', get_slurm_status)
        result = RUNNER.invoke(cli.slurm, ["status", "--run_name", "run", "--cluster_alias", self.cluster_alias])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Failed to get status' in result.output

    def test_slurm_status_query_fails(self, monkeypatch):
        """grid slurm status query fails"""
        def get_slurm_status(*args, **kwargs):
            return None

        monkeypatch.setattr(resolvers, 'get_slurm_status', get_slurm_status)
        result = RUNNER.invoke(cli.slurm, ["status", "--run_name", "run", "--cluster_alias", self.cluster_alias])
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to send status request' in result.output

    def test_slurm_status_succeeds(self, monkeypatch):
        """grid slurm status succeeds"""
        def get_slurm_status(*args, **kwargs):
            return {
                'success': True,
            }

        monkeypatch.setattr(resolvers, 'get_slurm_status', get_slurm_status)
        result = RUNNER.invoke(cli.slurm, ["status", "--run_name", "run", "--cluster_alias", self.cluster_alias])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Finished getting status from cluster' in result.output

    def test_slurm_submit_fails(self, tmp_path, monkeypatch):
        """grid slurm submit fails"""

        script = tmp_path / "script.py"
        script.write_text("")

        def train_slurm_script(*args, **kwargs):
            return {
                'success': False,
            }

        monkeypatch.setattr(resolvers, 'train_slurm_script', train_slurm_script)
        result = RUNNER.invoke(cli.slurm, [
            "train",
            script.as_posix(),
            "--cluster_alias",
            self.cluster_alias,
        ],
                               input='y\ny\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Unable to submit train task with error' in result.output

    def test_slurm_submit_fails_with_ex(self, tmp_path, monkeypatch):
        """grid slurm submit fails with exception"""

        script = tmp_path / "script.py"
        script.write_text("")

        def train_slurm_script(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'train_slurm_script', train_slurm_script)
        result = RUNNER.invoke(cli.slurm, [
            "train",
            script.as_posix(),
            "--cluster_alias",
            self.cluster_alias,
        ],
                               input='y\ny\n')
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to start training' in result.output

    def test_slurm_submit_succeeds(self, tmp_path, monkeypatch):
        """grid slurm submit succeeds"""

        script = tmp_path / "script.py"
        script.write_text("")

        def train_slurm_script(*args, **kwargs):
            return {
                'success': True,
            }

        monkeypatch.setattr(resolvers, 'train_slurm_script', train_slurm_script)
        result = RUNNER.invoke(cli.slurm, [
            "train",
            script.as_posix(),
            "--cluster_alias",
            self.cluster_alias,
        ],
                               input='y\ny\n')
        assert result.exit_code == 0
        assert not result.exception
        assert 'Run submitted!' in result.output

    def test_slurm_get_token_succeeds(self, tmp_path, monkeypatch):
        """grid slurm get-token succeeds"""
        def get_slurm_auth_token(*args, **kwargs):
            return {
                'success': True,
            }

        monkeypatch.setattr(resolvers, 'get_slurm_auth_token', get_slurm_auth_token)
        result = RUNNER.invoke(cli.slurm, ["get-token"])
        assert result.exit_code == 0
        assert not result.exception
        assert 'Generating token for grid-daemon use' in result.output

    def test_slurm_get_token_fails(self, monkeypatch):
        """grid slurm get-token fails"""
        def get_slurm_auth_token(*args, **kwargs):
            return {
                'success': False,
            }

        monkeypatch.setattr(resolvers, 'get_slurm_auth_token', get_slurm_auth_token)
        result = RUNNER.invoke(cli.slurm, ["get-token"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to create auth token.' in result.output

    def test_slurm_get_token_fails_with_ex(self, monkeypatch):
        """grid slurm get-token fails with exception"""
        def get_slurm_auth_token(*args, **kwargs):
            raise Exception()

        monkeypatch.setattr(resolvers, 'get_slurm_auth_token', get_slurm_auth_token)
        result = RUNNER.invoke(cli.slurm, ["get-token"])
        assert result.exit_code == 1
        assert result.exception
        assert 'Failed to create auth token.' in result.output



import click
from click.testing import CliRunner
from tests.utilities import create_local_schema_client
import ulid

from grid import cli
from grid.client import Grid
import grid.globals as env


class TestStatus:
    """Test case for the status command"""
    @classmethod
    def setup_class(cls):
        cls.runner = CliRunner()

        cls.credentials_path = 'tests/data/credentials.json'
        cls.test_run = 'manual-test-run'

        # Setup the global DEBUG flag to True.
        env.DEBUG = True

        #  Monkey patches the GraphQL client to read from a local schema.
        def monkey_patch_client(self):
            self.client = create_local_schema_client()

        #  skipcq: PYL-W0212
        Grid._init_client = monkey_patch_client

    def monkey_patched_method(self, *args, **kwargs):
        return True

    def monkey_patch_exception(self, *args, **kwargs):
        raise click.ClickException('test')

    def test_status_works(self, monkeypatch):
        """grid status does not raise an exception"""
        old_status = Grid.status
        monkeypatch.setattr(Grid, 'status', self.monkey_patched_method)

        result = self.runner.invoke(cli.status, [self.test_run])
        assert result.exit_code == 0
        assert not result.exception

        # Set attribute again.
        monkeypatch.setattr(Grid, 'status', old_status)

    def test_status_accepts_export_parameters(self, monkeypatch):
        """grid status accepts export parameters"""
        old_status = Grid.status
        monkeypatch.setattr(Grid, 'status', self.monkey_patched_method)

        for t in ['csv', 'json']:
            result = self.runner.invoke(cli.status, [self.test_run, '--export', t])
            assert result.exit_code == 0
            assert not result.exception

        # Set attribute again.
        monkeypatch.setattr(Grid, 'status', old_status)

    def test_status_run(self, monkeypatch):
        """grid status interactive nodes output toggles with the run param"""
        old_status = Grid.status
        monkeypatch.setattr(Grid, 'status', self.monkey_patched_method)

        for r in [self.test_run, '']:
            result = self.runner.invoke(cli.status, ['--follow', r])
            assert result.exit_code == 0
            assert not result.exception

        # Set attribute again.
        monkeypatch.setattr(Grid, 'status', old_status)

    def test_status_experiment_fails(self, monkeypatch):
        old_status = Grid.status
        monkeypatch.setattr(Grid, 'status', self.monkey_patched_method)

        experiment_id = str(ulid.new())
        experiment_name = "foo-bar-exp0"

        for experiment in (experiment_id, experiment_name):
            result = self.runner.invoke(cli.status, [experiment])
            assert result.exit_code == 2
            assert result.exception

        # Set attribute again.
        monkeypatch.setattr(Grid, 'status', old_status)



import os
from typing import Dict

import click

from grid.commands.git import execute_git_command
from grid.dependency_manager import CondaManager
from grid.dependency_manager import PipManager
import grid.globals as env

# TODO: Custom requirement files from user
PIP_REQUIREMENTS_FILE = "requirements.txt"
CONDA_REQUIREMENTS_FILE = "environment.yml"
DISABLE_ON_DEPENDENCY_FILE_EXIST = True


class DependencyMixin:
    def _serialize_dependencies(self, config: Dict):
        """
        Serialize dependency information to the corresponding source.

        Write python dependencies to the corresponding dependency listing
        (requirements.txt or environment.yml) and system dependencies to
        grid config (config.yml). Both write operations are actually offloaded
        downstream to the dependency manager object. Also fetches warnings
        from the dependency manager and echo to the user's terminal

        config:
            Grid config either read from the file provided by the user or the
            default generated one
        """
        self._deps_manager.write_spec()
        self._deps_manager.write_config(config)
        raise click.Abort()

    def _check_dependency_listing(self):
        """
        Evaluate if dependency listing is correct or not

        Determines the package manager employed by the user to specify
        application dependencies, setting up the appropriate interaction
        classes used to query one of the various package managers available
        as the system resolves dependency versions. Should no requirement file
        exist which specifies needed libraries+versions, this method provides
        the logic through which users are asked to select how they would like
        to proceed - automatic heuristic based generation or manual.

        Returns
        -------
        Return False if training can be continued, True if changes needs to be made
        """
        if env.IGNORE_WARNINGS:
            return False

        warning_str = click.style('WARNING', fg='yellow')
        docs_str = click.style('https://docs.grid.ai', fg='blue')
        conda_env = os.getenv("CONDA_DEFAULT_ENV")
        repo_root = execute_git_command(['rev-parse', '--show-toplevel'])
        # Assuming the requirements file is always at the repo root
        files = os.listdir(repo_root)

        if DISABLE_ON_DEPENDENCY_FILE_EXIST:
            if CONDA_REQUIREMENTS_FILE in files or PIP_REQUIREMENTS_FILE in files:
                return False

        if CONDA_REQUIREMENTS_FILE in files:
            if PIP_REQUIREMENTS_FILE in files:  # skipcq: PYL-R1705
                click.echo(f"{warning_str} Found both {CONDA_REQUIREMENTS_FILE} "
                           f"and {PIP_REQUIREMENTS_FILE}. Delete one of them if"
                           "you need us to run the dependency check. Ignoring for now")
                return False
            else:
                manager_class = CondaManager
        elif PIP_REQUIREMENTS_FILE in files:
            manager_class = PipManager
        elif conda_env:
            manager_class = CondaManager
        else:
            manager_class = PipManager

        # Returning False in-case of exceptions
        try:
            self._deps_manager = manager_class()
        except Exception as e:
            message = f"""

        {warning_str}

        We have encountered error while checking dependencies and dependency
        listing files (requirements.txt or environment.yml). Reach out to Grid
        Support at support@grid.ai with the error message below if you have
        questions

        "{e}"

        You can use the flag --ignore_warnings to skip this warning.
        See details at: {docs_str}

        Do you want to ignore the error and continue?
        """
            click.confirm(message, abort=True)
            return False

        # Returning if we find packages those are not specified in the
        # requirement listing or installed in the current environment
        missing = "\n        ".join(self._deps_manager.get_missing())
        if missing and not env.IGNORE_WARNINGS:
            message = f"""

        {warning_str}

        We found below packages being used in the source code but it is neither
        in your requirement listing (requirements.txt or environment.yml) nor
        installed in your current active environment.

        {missing}

        You can use the flag --ignore_warnings to skip this warning.
        See details at: {docs_str}

        Do you want to continue?
        """

            click.confirm(message, abort=True)
            return False  # returning if user continues

        message = f"""

        {warning_str}

        Listing of dependencies (requirements.txt or environment.yml) is
        incomplete/unavailable. This might lead to errors when running an Experiment
        in Grid. We can generate the dependency files for you by analysing your
        environment and source code. However, you'd still need to verify the files,
        commit and push it to your repository manually.

        Choose from below to continue

        [1] Generate dependency files and quit me from training
        [2] I don't care. Continue to train
        [3] Let me handle it. Exit!

        You can use the flag --ignore_warnings to skip this warning.
        See details at: {docs_str}

        """
        user_choice = "2"  # Ignore and continue to train
        if self._deps_manager.has_change:
            valid_choices = click.Choice(("1", "2", "3"))
            user_choice = click.prompt(message, show_default=False, default="1", type=valid_choices)
        if user_choice == "3":  # Ignore and exit
            raise click.Abort()
        # True if user decides to change deps and exit
        return True if user_choice == "1" else False  # skipcq: PYL-R1719



import ast
from shlex import split
from typing import Optional

import click
from gql import Client
from gql import gql
from rich.table import Table
import websockets

from grid.observables.base import BaseObservable
from grid.observables.base import style_status
from grid.observables.base import TASK_CLASS_MAPPING
from grid.utilities import get_experiment_duration_string
from grid.utilities import get_param_values


class Experiment(BaseObservable):
    def __init__(self, client: Client, identifier: str):
        self.client = client
        self.run_name = identifier

        super().__init__(client=client)

    def get_history(self, experiment_id: Optional[str] = None):
        """
        Parameters
        ----------
        experiment_id: Optional[str]
            Experiment ID
        """
        self.spinner.start()
        self.spinner.text = 'Getting Experiments ...'

        query = gql("""
        query (
            $runName: ID!
        ) {
            getExperiments(runName: $runName) {
                name
                status
                invocationCommands
                createdAt
                finishedAt
                commitSha
                run {
                    runId
                }
                startedRunningAt
            }
        }
        """)
        params = {'runName': self.run_name}
        try:
            self._get_task_run_status(run_name=self.run_name)
        except ValueError:
            return

        result = self.client.execute(query, variable_values=params)

        if not result['getExperiments']:
            click.echo(f'No experiments available for run "{self.run_name}"')
            return

        self.spinner.text = 'Done!'
        self.spinner.ok("✔")
        self.spinner.stop()
        table = self.render_experiments(result['getExperiments'])
        self.console.print(table)

    def get(self, experiment_id: Optional[str] = None):
        """
        Parameters
        ----------
        experiment_id: Optional[str]
            Experiment ID
        """
        self.spinner.start()
        self.spinner.text = 'Fetching experiment status ...'

        query = gql("""
        query (
            $runName: ID!
        ) {
            getExperiments(runName: $runName) {
                name
                status
                invocationCommands
                createdAt
                finishedAt
                commitSha
                run {
                    runId
                }
                startedRunningAt
            }
        }
        """)
        params = {'runName': self.run_name}
        try:
            self.spinner.text = 'Fetching experiment status ... done!'
            self._get_task_run_status(run_name=self.run_name)
            self.spinner.ok("✔")
        except ValueError:
            return
        finally:
            self.spinner.stop()

        result = self.client.execute(query, variable_values=params)

        experiments = result['getExperiments']

        if not experiments:
            click.echo(f'No experiments available for run "{self.run_name}"')
            return

        table = self.render_experiments(experiments)
        self.console.print(table)
        return result

    @staticmethod
    def render_experiments(experiments) -> Table:
        base_columns = [
            'Experiment',
            'Command',
            'Status',
            'Duration',
        ]
        if not experiments:
            return BaseObservable.create_table(columns=base_columns)

        command = experiments[0]['invocationCommands']
        toks = split(command)
        hparams = [tok.replace('--', '') for tok in toks if '--' in tok]

        table_columns = base_columns + hparams
        table = BaseObservable.create_table(columns=table_columns)

        for experiment in experiments:
            # Split hparam vals
            command = experiment['invocationCommands']
            base_command, *hparam_vals = get_param_values(command)
            # Get job duration - Since experiment started if running, since experiment created if queued
            duration_str = get_experiment_duration_string(created_at=experiment['createdAt'],
                                                          started_running_at=experiment['startedRunningAt'],
                                                          finished_at=experiment['finishedAt'])
            table.add_row(experiment['name'], base_command, experiment['status'], duration_str, *hparam_vals)
        return table

    def follow(self):
        """Follows a stream for a given Run."""
        self.spinner.text = 'Following Run details ...'
        self.spinner.start()

        #  Defines terminal states.
        terminal_state = ('finished', 'failed', 'cancelled')

        #  Get list of dependency IDs.
        dependencies = self._get_task_run_dependencies(run_name=self.run_name)

        # If no dependencies exists
        if not dependencies:
            raise click.ClickException("Run doesn't have dependencies. Nothing to follow.")

        dependency_ids = []
        dependency_statuses = []
        dependency_mapping = {}
        for dependency in dependencies:
            status = dependency['status']
            dependency_statuses.append(status)
            if status not in terminal_state:
                dependency_ids.append(dependency['taskId'])
                dependency_mapping[dependency['taskId']] = dependency

        #  Shows useful message to user.
        if all(s in terminal_state for s in dependency_statuses):
            self.spinner.ok("✔")
            self.spinner.stop()

            error_message = ''
            if any(s == 'failed' for s in dependency_statuses):
                error_message = f"""Use the following command to see errors in Run:
    grid status {self.run_name} --details"""

            # If an error was detected, raise an exception
            if error_message != '':
                click.echo()
                raise click.ClickException(f"""{error_message}""")

            # Get experiment statuses
            self.get()

            # Return here because there's no more scheduling steps to follow.
            return

        # If experiments are not in a terminal state,
        # open a websocket connection and consume messages.
        subscription = gql("""
        subscription GetTaskStream ($taskIds: [ID]!) {
            getTaskMessage(
                taskIds: $taskIds) {
                    taskId
                    message
                    timestamp
                    className
            }
        }
        """)

        params = {'taskIds': dependency_ids}

        #  Create a GraphQL subscription via websockect
        #  connection.
        try:
            stream = self.client.subscribe(subscription, variable_values=params)

            first_message = True
            for log in stream:

                #  Closes the spinner.
                if first_message:
                    self.spinner.text = 'Done!'
                    self.spinner.ok("✔")
                    first_message = False

                #  Patch a class name if not available with
                #  a placeholder.
                class_name = log['getTaskMessage']['className']
                if not class_name:
                    class_name = '-' * 10
                else:
                    class_name = TASK_CLASS_MAPPING[class_name]

                #  Extract task ID and status.
                task_id = log['getTaskMessage']['taskId']
                task_status = dependency_mapping[task_id]['status']

                #  Print each line to terminal.
                styled_class_name = style_status(class_name, task_status)
                message = log['getTaskMessage']['message']
                click.echo(f'[{styled_class_name}] {message}')

        # If connection is suddenly closed, indicate that a
        # known error happened.
        except websockets.exceptions.ConnectionClosedError:
            self.spinner.fail("✘")
            self.spinner.stop()
            raise click.ClickException('Could not fetch log data.')

        except websockets.exceptions.ConnectionClosedOK:
            self.spinner.fail("✘")
            self.spinner.stop()
            raise click.ClickException('Could not continue fetching log stream.')

        #  Raise any other errors that the backend may raise.
        except Exception as e:  # skipcq: PYL-W0703
            self.spinner.fail("✘")
            self.spinner.stop()
            if 'Server error:' in str(e):
                e = str(e).replace('Server error: ', '')[1:-1]

            message = ast.literal_eval(str(e))['message']
            raise click.ClickException(message)



from typing import List

import click

from grid.client import Grid


@click.group()
def cancel() -> None:
    pass


@cancel.command()
@click.argument('experiment_ids', type=str, required=True, nargs=-1)
def experiment(experiment_ids: List[str]):
    client = Grid()
    for experiment in experiment_ids:
        client.cancel(experiment_name=experiment)


@cancel.command()
@click.argument('run_ids', type=str, required=True, nargs=-1)
def run(run_ids: List[str]):
    client = Grid()
    for run in run_ids:
        client.cancel(run_name=run)



from pathlib import Path
import sys

from pkg_resources import Requirement
import pytest

from grid import dependency_manager
from grid.dependency_manager import _conda2parsed_pypi
from grid.dependency_manager import _get_rule_maps
from grid.dependency_manager import _get_system_deps
from grid.dependency_manager import _guess_valid_name
from grid.dependency_manager import _source2parsed_pypi
from grid.dependency_manager import CondaManager
from grid.dependency_manager import CondaPackage
from grid.dependency_manager import NonComparableError
from grid.dependency_manager import PipManager
from grid.dependency_manager import PipPackage
from grid.dependency_manager import yaml


class TestRule:
    def test_get_rule_map(self):
        rule_labels = {"pip_name_mapping", "conda_name_mapping", "system"}
        rules = _get_rule_maps()

        # verifying rule labels
        assert len(rule_labels.symmetric_difference(rules.keys())) == 0

        # verifying types for each labels
        assert isinstance(rules["pip_name_mapping"], dict)
        key = list(rules["pip_name_mapping"].keys())[0]
        assert isinstance(rules["pip_name_mapping"][key], str)

        assert isinstance(rules["conda_name_mapping"], dict)
        key = list(rules["conda_name_mapping"].keys())[0]
        assert isinstance(rules["conda_name_mapping"][key], str)

        assert isinstance(rules["system"], dict)
        key = list(rules["system"].keys())[0]
        assert isinstance(rules["system"][key], list)

    def test_hyphen_in_rules(self):
        rules = _get_rule_maps()
        for name, value in rules["pip_name_mapping"].items():
            assert "-" not in name
            assert "-" not in value

        for value in rules["conda_name_mapping"].values():
            assert "-" not in value

        for name in rules["system"]:
            assert "-" not in name

    def test_get_system_deps_missing_rule(self):
        # Should return an empty list if a rule is missing
        assert _get_system_deps("dummy_name") == []

    def test_conda2parsed_pypi_missing_rule(self):
        # Should return the same name if a name is missing
        assert _conda2parsed_pypi("dummy_name") == "dummy_name"

    def test_source2parsed_pypi_name_for_missing_rule(self):
        assert _source2parsed_pypi("dummy-name") == "dummy_name"


class TestDifferentRequirementsStringForPIP:
    # these strings are tested in the test cases. Keeping here as
    # documentation
    strings = [
        "dummy-name==0.1.0",
        "dummy@ git+https://github.com/dummy/dummy",
        "dummy>0.1.0,<=1.0.0",
        "dummy-name",
        "pkg==1.4.0;python_version > '3.5'",
        "pkg==1.4.0 -i https://someurl.com/repo",
        "pkg==1.4.0 --index-url https://someurl.com/repo",
        "git+https://github.com/dummy/dummy.git",
        "git+https://github.com/dummy/dummy",
        "https://repo.com/dummy.tar.gz",
        "-e git+https://github.com/dummy/dummy.git",
        ".",
        "/some/path/outside/of/cwd",
        "path/to/wheel.whl",
        "-e file:///path/to/package",
        " pkg==1.4.0",  # string starting with space
        " git+https://github.com/dummy/dummy",  # url starting with space
        "--invalid-option pkg==1.4.0",
        "dummy @ file:///some/local/path",
        "dummy @ /some/local/path",
        "pkg==1.4.0; python_version < '2.7'",
        "pkg==1.4.0;python_version < '2.7'",
        "git+https://github.com/dummy/dummy; python_version < '2.7'",
        "git+https://github.com/dummy/dummy;python_version < '2.7'",  # invalid
        ".==1.4",  # invalid
        "valid/path; python_version < 1.4",
        "valid/path;python_version < 1.4",
        "-e file:///path/to/package; python_version < 1.4",
        "git+https://github.com/dummy/dummy==1.4",  # invalid
        "invalid:requirement$string"
    ]

    def test_name_and_version(self):
        spec = "dummy-name==0.1.0"
        pkg = PipPackage.from_string(spec)
        assert pkg._requirement_obj.url is None
        assert pkg.name == "dummy-name"
        assert pkg.parsed_name == "dummy_name"
        assert pkg._requirement_obj.specs == [("==", "0.1.0")]
        assert pkg._original_string is None
        assert str(pkg) == spec

    def test_name_and_git_url(self):
        spec = "dummy@ git+https://github.com/dummy/dummy"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg._requirement_obj.url == "git+https://github.com/dummy/dummy"
        # This situation never occurs since our env package won't have version
        # TODO: test the above assumption
        assert PipPackage.from_string("dummy==0.1.0") != pkg  # noqa

    @pytest.mark.skipif(sys.version_info < (3, 8),
                        reason="spec of the format `name@ path` is not "
                        "supported by pkg_resources in 3.7")
    def test_name_and_file_url(self):
        spec = "dummy@ file:///some/local/path"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg != PipPackage.from_string("dummy")  # noqa

    @pytest.mark.skipif(sys.version_info >= (3, 8),
                        reason="spec of the format `name@ path` is not "
                        "supported by pkg_resources in 3.7")
    def test_name_and_file_url_for_37(self):
        spec = "dummy@ file:///some/local/path"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        with pytest.raises(NonComparableError):
            pkg != PipPackage.from_string("dummy")  # noqa

    def test_name_and_path(self, monkeypatch):
        spec = "dummy@ /some/local/path"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert len(pkg.name) == 5
        monkeypatch.setattr(dependency_manager.Path, "exists", lambda x: True)
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "dummy"

    def test_path_and_specifier(self):
        spec = ".==1.4"
        with pytest.raises(RuntimeError, match="Error while parsing"):
            PipPackage.from_string(spec)

    def test_url_and_specifier(self):
        spec = "git+https://github.com/dummy/dummy==1.4"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec  # noqa
        # Not comparable since pkg._requirement_obj is None
        with pytest.raises(NonComparableError):
            pkg == PipPackage.from_string("dummy==1.4")  # noqa
        with pytest.raises(NonComparableError):
            PipPackage.from_string("dummy==1.4") == pkg  # noqa

    def test_name_with_multiple_specifiers(self):
        spec = "dummy<=1.0.0,>0.1.0"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec

    def test_name_version_and_marker(self):
        spec = "pkg==1.4.0; python_version < '2.7'"
        pkg = PipPackage.from_string(spec)
        assert pkg._requirement_obj == Requirement.parse(spec)
        spec = "pkg==1.4.0;python_version < '2.7'"
        pkg = PipPackage.from_string(spec)
        assert pkg._requirement_obj == Requirement.parse(spec)

    def test_url_and_marker(self):
        spec = "git+https://github.com/dummy/dummy; python_version < '2.7'"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg.name == "dummy"

        # URL can't parse correctly hence name is uuid
        spec = "git+https://github.com/dummy/dummy;python_version < '2.7'"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        # not a valid name but we are ignoring this innocuous error
        assert pkg.name == "dummy;python_version < '2"

    def test_path_and_marker(self, monkeypatch):
        monkeypatch.setattr(dependency_manager.Path, "exists", lambda x: True)
        spec = "valid/path/dummy; python_version < 1.4"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg.name == "dummy"
        spec = "valid/path/dummy;python_version < 1.4"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg.name == "dummy"

    def test_just_name(self):
        spec = "dummy-name"
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec

    def test_name_and_markers(self):
        spec = 'pkg==1.4.0; python_version > "3.5"'
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert PipPackage.from_string("pkg==1.4.0") == pkg

    def test_name_and_option(self):
        spec = "pkg==1.4.0 -i https://someurl.com/repo"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "pkg"
        assert str(pkg) == spec
        with pytest.raises(NonComparableError):
            PipPackage.from_string("pkg==100.4.0") == pkg  # noqa

    def test_name_and_long_option(self):
        spec = "pkg==1.4.0 --index-url https://someurl.com/repo"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "pkg"
        assert str(pkg) == spec
        with pytest.raises(NonComparableError):
            PipPackage.from_string("pkg==100.4.0") == pkg  # noqa

        spec = "pkg==1.4.0 --index-url https://someurl.com/repo -f https://anotherurl.com/"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "pkg"
        assert str(pkg) == spec

    def test_git_url(self):
        spec = "git+https://github.com/dummy/dummy.git"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "dummy"
        spec = "git+https://github.com/dummy/dummy"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "dummy"

    def test_normal_url(self):
        spec = "https://repo.com/dummy.tar.gz"
        pkg = PipPackage.from_string(spec)
        assert pkg.name == "dummy"
        assert str(pkg) == spec

    def test_editable_git_url(self):
        spec = "-e git+https://github.com/dummy/dummy.git"
        with pytest.raises(RuntimeError, match="editable requirements"):
            PipPackage.from_string(spec)

    def test_editable_file_url(self):
        spec = "-e file:///path/to/package"
        with pytest.raises(RuntimeError, match="editable requirements"):
            PipPackage.from_string(spec)

    def test_editable_file_url_with_marker(self):
        spec = "-e file:///path/to/package; python_version < 1.4"
        with pytest.raises(RuntimeError, match="editable requirements"):
            PipPackage.from_string(spec)

    def test_current_directory(self):
        spec = "."
        pkg = PipPackage.from_string(spec)
        assert pkg.name == 'grid-cli'
        with pytest.raises(NonComparableError):
            PipPackage.from_string("grid-cli==0.1") == pkg  # noqa

    def test_outside_directory(self):
        spec = "/some/path/outside/of/cwd"
        with pytest.raises(RuntimeError, match="not resolve"):
            PipPackage.from_string(spec)

    def test_inside_directory(self, monkeypatch):
        spec = "path/to/wheel.whl"
        with pytest.raises(RuntimeError, match="Error while parsing"):
            PipPackage.from_string(spec)
        monkeypatch.setattr(dependency_manager.Path, "exists", lambda x: True)
        pkg = PipPackage.from_string(spec)
        assert str(pkg) == spec
        assert pkg.name == "wheel"

        with pytest.raises(NonComparableError):
            pkg != PipPackage.from_string("pkg")  # noqa

    def test_string_starts_with_space(self):
        spec = " pkg==1.4.0"
        pkg = PipPackage.from_string(spec)
        assert PipPackage.from_string(spec.strip()) == pkg
        spec = " git+https://github.com/dummy/dummy"
        pkg = PipPackage.from_string(spec)
        with pytest.raises(NonComparableError):
            PipPackage.from_string(spec.strip()) == pkg  # noqa

    def test_invalid_option_at_starting(self):
        spec = "--invalid-option pkg==1.4.0"
        with pytest.raises(RuntimeError, match="Error while parsing"):
            PipPackage.from_string(spec)

    def test_invalid_string(self):
        spec = "invalid:requirement$string"
        with pytest.raises(RuntimeError, match="Valid URL schemes"):
            PipPackage.from_string(spec)


class TestPackageContainer:
    def test_pip_package_creation_from_requirement(self):
        req = Requirement.parse("dummy-name==0.1.0")
        pkg = PipPackage.from_requirement(req)
        assert pkg.name == "dummy-name"
        assert pkg.parsed_name == "dummy_name"

    def test_pip_package_from_distribution(self, monkeypatch):
        class DummyDistribution:
            project_name = "dummy-name"
            as_requirement = lambda x: Requirement.parse("dummy-name==0.1.0")  # noqa

        monkeypatch.setattr(dependency_manager, "_is_dist_editable", lambda x: True)
        with pytest.raises(RuntimeError, match="editable installation"):
            PipPackage.from_distribution(DummyDistribution())  # noqa

        monkeypatch.setattr(dependency_manager, "_is_dist_editable", lambda x: False)
        monkeypatch.setattr(dependency_manager, "_get_distribution_url", lambda x: None)
        pkg1 = PipPackage.from_distribution(DummyDistribution())  # noqa
        assert pkg1.name == "dummy-name"
        assert str(pkg1) == "dummy-name==0.1.0"

        monkeypatch.setattr(dependency_manager, "_get_distribution_url", lambda x: "https://github.com/org/repo")
        pkg2 = PipPackage.from_distribution(DummyDistribution())  # noqa
        assert str(pkg2) == "dummy-name@ https://github.com/org/repo"
        # TODO: self._requirement_obj == other._requirement_obj is
        #  True but it shouldn't be
        assert pkg1 == pkg2  # noqa

    def test_pip_package_equality(self):
        # More granular equality checks are done in
        # TestDifferentRequirementsStringForPIP
        pkg1 = PipPackage.from_string("dummy-name==0.1.0")
        pkg2 = PipPackage.from_string("dummy-name==0.2.0")
        assert pkg1 != pkg2

        # Pip package and conda package
        pkg2 = CondaPackage.from_string("dummy=hash=version")
        assert pkg1 != pkg2

        # Underscore
        pkg1 = PipPackage.from_string("dummy-name")
        pkg2 = PipPackage.from_string("dummy_name")
        assert pkg1 == pkg2

        # version inclusion
        pkg1 = PipPackage.from_string("dummy-name>1.0,<2.0")
        pkg2 = PipPackage.from_string("dummy_name==1.5")
        assert pkg1 == pkg2
        assert pkg2 == pkg1

        # Different version
        pkg3 = PipPackage.from_string("dummy_name==2.5")
        assert pkg1 != pkg3

        pkg4 = PipPackage.from_string("dummy-name>1.5,<2.0")
        assert pkg1 != pkg4

        pkg1 = PipPackage.from_string("dummy-name==2.0")
        pkg2 = PipPackage.from_string("dummy-name")
        pkg3 = PipPackage.from_string("dummy-name>=")
        # first specifier has empty version
        pkg4 = PipPackage.from_string("dummy-name>=, <3.0")
        assert pkg1 != pkg2
        assert pkg1 != pkg3
        assert pkg1 == pkg4

        # different parse_name - should never happen in real world
        pkg1 = PipPackage.from_string("dummy-name==2.0")
        pkg2 = PipPackage.from_string("dummy-name==2.0")
        pkg2.parsed_name = "wrong_parsed_name"
        with pytest.raises(NonComparableError):
            pkg1 != pkg2  # noqa

    def test_conda_package_creation(self):
        pkg = CondaPackage(name="dummy-name", version="0.1.0", build="py1000")
        assert pkg.parsed_name == "dummy_name"

        pkg = CondaPackage.from_string("dummy-name=0.1.0=py1000")
        assert pkg.build == "py1000"

        pkg = CondaPackage.from_string("dummy==0.1")
        assert pkg.name == "dummy"
        assert pkg.build is None
        assert pkg.version == "0.1"

        pkg = CondaPackage.from_string("dummy==0.1==py1000")
        assert pkg.name == "dummy"
        assert pkg.build == "py1000"
        assert pkg.version == "0.1"

        pkg = CondaPackage.from_string("dummy==0.1=py1000")
        assert pkg.name == "dummy"
        assert pkg.build == "py1000"
        assert pkg.version == "0.1"

        pkg = CondaPackage.from_string("dummy=0.1.0")
        assert pkg.name == "dummy"
        assert pkg.build is None
        assert pkg.version == "0.1.0"

        pkg = CondaPackage.from_string("dummy")
        assert pkg.name == "dummy"
        assert pkg.version is None
        assert pkg.build is None

        with pytest.raises(RuntimeError, match="ParsingError"):
            CondaPackage.from_string("dummy=0.1.0=py1000=fake")

    def test_conda_package_to_string(self):
        pkg = CondaPackage(name="dummy-name", version="0.1.0", build="py1000")
        assert str(pkg) == "dummy-name=0.1.0=py1000"
        pkg = CondaPackage(name="dummy-name", version=None, build=None)
        assert str(pkg) == "dummy-name"
        pkg = CondaPackage(name="dummy-name", version="0.1.0", build=None)
        assert str(pkg) == "dummy-name=0.1.0"

    def test_conda_package_equality(self):
        pkg1 = CondaPackage(name="dummy-name", version="0.1.0", build="py100")
        pkg2 = CondaPackage(name="dummy-name", version="0.1.0", build="py100")
        pkg3 = PipPackage.from_string("dummy-name")
        assert pkg1 == pkg2
        assert pkg1 != pkg3

        pkg1.build = "another build"
        assert pkg1 == pkg2  # still equal - build doesn't have any effect

        pkg1.version = "0.2.0"
        assert pkg1 != pkg2


class TestDepsManagerBaseClass:
    def test_read_file_deps_with_no_requirement_file(self, monkeypatch):
        monkeypatch.setattr(PipManager, "_setup", lambda x: None)
        pm = PipManager()
        pm.req_file = None
        monkeypatch.setattr(pm, "req_file", Path("filedoesnotexist"))
        assert pm._read_file_deps() == {}

    def test_get_missing(self, monkeypatch):
        source_deps = ["pkg_a", "pkg_b", "pkg_c"]

        def test_setup(self):
            self.env_deps = {"pkg_z": object(), "pkg_a": object(), "pkg_b": object()}
            self.file_deps = {"pkg_b": object(), "pkg_c": object(), "pkg_d": object()}
            self._source_deps_list = source_deps

        # No missing keys
        monkeypatch.setattr(PipManager, "_setup", test_setup)
        pm = PipManager()
        assert set() == pm.get_missing()

        # adding missing keys
        source_deps.append("pkg_missing")
        assert {"pkg_missing"} == pm.get_missing()

    def test_has_change(self, monkeypatch):
        source_deps = ["pkg_a", "pkg_b"]

        class DummyFile:
            def __init__(self):
                self.exists = lambda: False

        dummyfile = DummyFile()

        def test_setup(self):
            self.req_file = dummyfile
            self.env_deps = {"pkg_z": object(), "pkg_a": object(), "pkg_b": object()}
            self.file_deps = {"pkg_b": object(), "pkg_c": object(), "pkg_d": object()}
            self._source_deps_list = source_deps

        # No requirement file exist
        monkeypatch.setattr(PipManager, "_setup", test_setup)
        pm = PipManager()
        assert pm.has_change is True

        # requirement file exist but keys from source is not present in file deps
        dummyfile.exists = lambda: True
        assert pm.has_change is True

        # keys from source is not in env deps which is OK (pkg_d)
        # env deps package != file deps package (pkg_b)
        source_deps.clear()
        source_deps.append("pkg_d")
        source_deps.append("pkg_b")
        assert pm.has_change is True

    def test_collate_final(self, monkeypatch):
        source_deps = ["pkg_a", "pkg_b", "pkg_c", "pkg_d"]
        obja, objb, objc, objd = object(), object(), object(), object()

        def test_setup(self):
            self.env_deps = {"pkg_d": objd}
            self.file_deps = {"pkg_c": objc, "pkg_b": objb, "pkg_a": obja}
            self._source_deps_list = source_deps

        # Adding missing deps & keeping the order of elements in file deps
        monkeypatch.setattr(PipManager, "_setup", test_setup)
        pm = PipManager()
        assert pm.collate_final() == [objc, objb, obja, objd]

        # Updating deps from env if env deps is not same as file deps
        newobjc = object()
        pm.env_deps = {"pkg_d": objd, "pkg_c": newobjc}
        assert pm.collate_final() == [newobjc, objb, obja, objd]

    def test_write_config(self, monkeypatch):
        source_deps = ["pkg_a", "pkg_b", "pkg_c", "pkg_d"]
        # monkeypatching safedump to ease testing
        yamlout = []
        monkeypatch.setattr(yaml, "safe_dump", lambda c, f: yamlout.append((c, f)))

        def test_setup(self):
            self._source_deps_list = source_deps

        # No system deps found
        monkeypatch.setattr(PipManager, "_setup", test_setup)
        pm = PipManager()
        assert pm.write_config({"compute": {}}) is None

        # Empty config
        monkeypatch.setattr(dependency_manager, "_get_system_deps", lambda x: [x])
        with pytest.raises(KeyError, match="compute"):
            pm.write_config({})

        # multiple system deps
        pm.write_config({"compute": {}})
        assert yamlout[0][0] == {'compute': {'actions': {'on_image_build': ['apt install -y pkg_a pkg_b pkg_c pkg_d']}}}

        configfile = yamlout[0][1]
        configfile.close()
        Path(configfile.name).unlink()

        yamlout.clear()
        pm.write_config({"compute": {"actions": {"on_image_build": []}}})
        assert yamlout[0][0] == {'compute': {'actions': {'on_image_build': ['apt install -y pkg_a pkg_b pkg_c pkg_d']}}}

        configfile = yamlout[0][1]
        configfile.close()
        Path(configfile.name).unlink()


class TestPipManager:
    def test_read_file_deps_success(self, monkeypatch):
        class DummyFile:
            def __init__(self):
                self.exists = lambda: True
                self.read_text = lambda: "pkg-a==0.2\npkg-b"

        monkeypatch.setattr(PipManager, "_setup", lambda x: None)
        pm = PipManager()
        pm.req_file = None
        dummyfile = DummyFile()
        monkeypatch.setattr(pm, "req_file", dummyfile)
        final = {"pkg_a": PipPackage.from_string("pkg-a==0.2"), "pkg_b": PipPackage.from_string("pkg-b")}
        assert pm._read_file_deps() == final

    def test_read_file_deps_with_invalid_file(self, monkeypatch):
        class DummyFile:
            def __init__(self):
                self.exists = lambda: True
                self.read_text = lambda: "channel:  - pytorch\n"

        monkeypatch.setattr(PipManager, "_setup", lambda x: None)
        pm = PipManager()
        pm.req_file = None
        dummyfile = DummyFile()
        monkeypatch.setattr(pm, "req_file", dummyfile)
        with pytest.raises(RuntimeError, match="Valid URL schemes"):
            pm._read_file_deps()

    def test_get_default_file(self, monkeypatch):
        monkeypatch.setattr(PipManager, "_setup", lambda x: None)
        pm = PipManager()
        assert Path("requirements.txt") == pm._get_default_req_file()


class TestCondaManager:
    def test_read_file_deps_success(self, monkeypatch):
        class DummyFile:
            def __init__(self):
                self.exists = lambda: True
                self.read_text = lambda: 'channels:\n  - pytorch\n  - defaults\ndependencies:\n  - pkg-b=20.1.0=py38\n  - pkg-a=1.10=py_0\n'

        monkeypatch.setattr(CondaManager, "_setup", lambda x: None)
        cm = CondaManager()
        cm.req_file = None
        dummyfile = DummyFile()
        monkeypatch.setattr(cm, "req_file", dummyfile)
        final = {
            "pkg_b": CondaPackage.from_string("pkg-b=20.1.0=py38"),
            "pkg_a": CondaPackage.from_string("pkg-a=1.10=py_0")
        }
        assert cm._read_file_deps() == final

    def test_read_file_deps_with_invalid_file(self, monkeypatch):
        class DummyFile:
            def __init__(self):
                self.exists = lambda: True
                self.read_text = lambda: "pkg-a==0.2\npkg-b"

        monkeypatch.setattr(CondaManager, "_setup", lambda x: None)
        pm = CondaManager()
        pm.req_file = None
        dummyfile = DummyFile()
        monkeypatch.setattr(pm, "req_file", dummyfile)
        with pytest.raises(RuntimeError):
            # generic exception since different python version raises
            # different exception
            pm._read_file_deps()

    def test_get_default_file(self, monkeypatch):
        monkeypatch.setattr(CondaManager, "_setup", lambda x: None)
        pm = CondaManager()
        assert Path("environment.yml") == pm._get_default_req_file()


def test_get_valid_stem():
    with pytest.raises(RuntimeError, match="not portable"):
        _guess_valid_name("file://path/to/dir")
    with pytest.raises(RuntimeError, match="Valid URL schemes are"):
        _guess_valid_name("invalid://path/to/dir")
    with pytest.raises(RuntimeError, match="Error while parsing"):
        _guess_valid_name("invalid;path--with==random+char#%^")
    assert "path" == _guess_valid_name("https://github.com/org/path@branch")
    assert "path" == _guess_valid_name("https://github.com/org/path")
    assert "name" == _guess_valid_name("name @ path")



"""
Create tar file from folder
"""
from dataclasses import dataclass
import math
import os
import subprocess
from typing import Optional, Tuple


def get_dir_size_and_count(source_dir: str, prefix: Optional[str] = None) -> Tuple[int, int]:
    """
    Get size and file count of a directory

    Parameters
    ----------
    source_dir: str
        Directory path

    Returns
    -------
    Tuple[int, int]
        Size in megabytes and file count
    """
    size = 0
    count = 0
    for root, _, files in os.walk(source_dir, topdown=True):
        for f in files:
            if prefix and not f.startswith(prefix):
                continue

            full_path = os.path.join(root, f)
            size += os.path.getsize(full_path)
            count += 1

    return (size, count)


@dataclass
class TarResults:
    """
    This class holds the results of running tar_path.

    Attributes
    ----------
    before_size: int
        The total size of the original directory files in bytes
    after_size: int
        The total size of the compressed and tarred split files in bytes
    """
    before_size: int
    after_size: int


def get_split_size(total_size: int, minimum_split_size: int = 1024 * 1000 * 20, max_split_count: int = 10000) -> int:
    """
    Calculate the split size we should use to split

    Parameters
    ----------
    minimum_split_size: int
        The minimum split size to use
    max_split_count: int
        The maximum split count
    total_size: int
        Total size of the file to split

    Returns
    -------
    int
        Split size
    """
    split_size = minimum_split_size
    split_count = math.ceil(total_size / split_size)
    if split_count > max_split_count:
        # Adjust the split size based on max split count
        split_size = math.ceil(total_size / max_split_count)

    return split_size


def tar_path(source_path: str, target_file: str, compression: bool = False) -> TarResults:
    """
    Create tar from directory using `tar`

    Parameters
    ----------
    source_path: str
        Source directory or file
    target_file
        Target tar file
    compression: bool, default False
        Enable compression, which is disabled by
        default.

    Returns
    -------
    TarResults
        Results that holds file counts and sizes
    """
    if os.path.isdir(source_path):
        before_size, _ = get_dir_size_and_count(source_path)
    else:
        before_size = os.path.getsize(source_path)

    # Only add compression when users explicitly request it.
    # We do this because it takes too long to compress
    # large datastores.
    tar_flags = '-cvf'
    if compression:
        tar_flags = '-zcvf'
    if os.path.isdir(source_path):
        command = f"tar -C {source_path} {tar_flags} {target_file} ./"
    else:
        abs_path = os.path.abspath(source_path)
        parent_dir = os.path.dirname(abs_path)
        base_name = os.path.basename(abs_path)
        command = f"tar -C {parent_dir} {tar_flags} {target_file} {base_name}"

    subprocess.check_call(command,
                          stdout=subprocess.DEVNULL,
                          stderr=subprocess.DEVNULL,
                          shell=True,
                          env={
                              "GZIP": "-9",
                              "COPYFILE_DISABLE": "1"
                          })

    after_size = os.stat(target_file).st_size
    return TarResults(before_size=before_size, after_size=after_size)
